from unittest import TestCase
from musicscore.musicxml.types.complextypes.notations import Slur


class Test(TestCase):
    def test_1(self):
        slur = Slur(type='start')
        result = '''<slur number="1" type="start"/>
'''
        self.assertEqual(slur.to_string(), result)



from lxml import etree as et

import musicscore.musicxml.elements.timewise as timewise
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treepagestyle import TreePageStyle
from musicscore.musictree.treepart import TreePart
from musicscore.musictree.treescorepart import TreeScorePart
from musicscore.musicxml.elements import partwise
from musicscore.musicxml.elements.barline import Barline, BarStyle
from musicscore.musicxml.elements.scoreheader import PartList, Credit, Defaults
from musicscore.musicxml.types.complextypes.appearance import LineWidth
from musicscore.musicxml.types.complextypes.credit import CreditType, CreditWords
from musicscore.musicxml.types.complextypes.defaults import Appearance, WordFont
from musicscore.musicxml.types.complextypes.encoding import Supports
from musicscore.musicxml.types.complextypes.identification import Encoding
from musicscore.musicxml.types.complextypes.scorepart import Identification


class TreeScoreTimewise(timewise.Score):
    """"""

    _auto_part_number = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._part_list = self.add_child(PartList())
        self.version = '3.0'
        self._tuplet_line_width = None
        self._finished = False
        self._pre_quantized = False
        self._quantized = False
        self._post_quantized = False
        self._max_division = None
        self._forbidden_divisions = None
        self._page_style = TreePageStyle(score=self, **kwargs)
        self._accidental_mode = 'normal'
        self._title = None
        self._subtitle = None
        self._composer = None

        self._identifications_added = False

    @property
    def title(self):
        return self._title

    @property
    def subtitle(self):
        return self._subtitle

    @property
    def composer(self):
        return self._composer

    @property
    def defaults(self):
        try:
            return self.get_children_by_type(Defaults)[0]
        except IndexError:
            return None

    def make_defaults(self):
        if self.defaults is None:
            self.add_child(Defaults())
        else:
            raise Exception('defaults already exists')

    @property
    def word_font(self):
        try:
            return self.defaults.get_children_by_type(WordFont)[0]
        except (AttributeError, IndexError):
            return None

    def make_word_font(self, **kwargs):
        if self.defaults is None:
            self.make_defaults()
        if self.word_font is None:
            self.defaults.add_child(WordFont(**kwargs))
        else:
            raise Exception('word-font already exists')

    @property
    def tuplet_line_width(self):
        return self._tuplet_line_width

    @tuplet_line_width.setter
    def tuplet_line_width(self, val):

        if not isinstance(val, float):
            raise TypeError('tuplet_line_width.value must be of type float not{}'.format(type(val)))
        self._tuplet_line_width = val

        if self.defaults is None:
            self.make_defaults()

        try:
            appearance = self.defaults.get_children_by_type(Appearance)[0]
        except IndexError:
            appearance = self.defaults.add_child(Appearance())

        try:
            line_width = [lw for lw in appearance.get_children_by_type(LineWidth) if lw.type == 'tuplet bracket'][0]
        except IndexError:
            line_width = appearance.add_child(LineWidth(type='tuplet bracket'))

        line_width.value = val

    def add_word_font(self, **kwargs):
        if self.defaults is None:
            self.make_defaults()

        self.defaults

    @property
    def max_division(self):
        return self._max_division

    @max_division.setter
    def max_division(self, value):
        if value is not None and not isinstance(value, int):
            raise TypeError('max_division.value must be None or of type int not {}'.format(type(value)))

        self._max_division = value

    @property
    def accidental_mode(self):
        return self._accidental_mode

    @accidental_mode.setter
    def accidental_mode(self, value):
        self._accidental_mode = value

    @property
    def forbidden_divisions(self):

        return self._forbidden_divisions

    @forbidden_divisions.setter
    def forbidden_divisions(self, value):
        if value is not None:
            for x in value:
                if not isinstance(x, int):
                    raise TypeError('forbidden_division must be of type int not{}'.format(type(value)))

        self._forbidden_divisions = value

    @property
    def page_style(self):
        return self._page_style

    def recalculate_x(self):
        for credit in self.get_children_by_type(Credit):
            for credit_words in credit.get_children_by_type(CreditWords):
                x_factor = self.page_style.page_width.value / self.page_style.previous_page_width_value
                credit_words.default_x *= x_factor

    def recalculate_y(self):
        for credit in self.get_children_by_type(Credit):
            for credit_words in credit.get_children_by_type(CreditWords):
                y_factor = self.page_style.page_height.value / self.page_style.previous_page_height_value
                credit_words.default_y *= y_factor

    def _generate_score_part(self):
        id_ = 'p' + str(self._auto_part_number)
        self._auto_part_number += 1
        return TreeScorePart(id=id_)

    def get_score_parts(self):
        return self._part_list.get_children_by_type(TreeScorePart)

    def get_score_part(self, id):
        for score_part in self.get_score_parts():
            if score_part.id == id:
                return score_part
        return None

    def add_score_part(self, score_part):
        score_part.parent_score = self
        return self._part_list.add_child(score_part)

    @property
    def number_of_parts(self):
        return len(self._part_list.get_children_by_type(TreeScorePart))

    def add_part(self, instrument=None):
        new_score_part = self._generate_score_part()
        new_score_part.instrument = instrument
        self.add_score_part(new_score_part)

        for measure in self.get_children_by_type(TreeMeasure):
            part = new_score_part.add_part()
            measure.add_child(part)
            if measure.barline_style:
                bl = part.add_child(Barline())
                bl.add_child(BarStyle(measure.barline_style))

    def add_measure(self, measure=None):
        new_measure = self._set_new_measure(measure)

        for score_part in self.get_score_parts():
            p = score_part.add_part()
            new_measure.add_child(p)
        return self.add_child(new_measure)

    def _set_new_measure(self, measure):
        if measure is None:
            new_measure = TreeMeasure()
            measures = self.get_children_by_type(TreeMeasure)
            if measures:
                new_measure.time = measures[-1].time.__copy__()
            else:
                new_measure.time.show = True
        else:
            new_measure = measure
        return new_measure

    def _add_identifications(self):
        identification = self.add_child(Identification())
        encoding = identification.add_child(Encoding())
        encoding.add_child(Supports(attribute='new-page', element='print', type_='yes', value_='yes'))
        encoding.add_child(Supports(attribute='new-system', element='print', type_='yes', value_='yes'))
        self._identifications_added = True

    def set_time_signatures(self, quarter_durations=None, times=None, barline_style=None):
        global current_time
        if self.get_children_by_type(TreeMeasure):
            raise Exception('for setting time signatures score should be empty')

        def make_measure(duration=None):
            def get_time():
                if duration % 1 == 0:
                    time = (int(duration), 4)
                    return time
                elif (duration * 2) % 1 == 0:
                    time = (int(duration * 2), 8)
                    return time
                else:
                    raise ValueError('duration {} is not dividable'.format(duration))

            if not duration:
                time = (4, 4)
            else:
                time = get_time()

            return TreeMeasure(time)

        if not times:
            times = {}

        current_time = (4, 4)

        def set_times(current_measure_number, duration):
            global current_time
            duration = float(duration)
            remaining_duration = duration

            while True:
                if current_measure_number in times.keys():
                    current_time = times[current_measure_number]

                measure = TreeMeasure(time=current_time)

                measure_duration = measure.quarter_duration
                if measure_duration > remaining_duration:
                    measure = make_measure(duration=remaining_duration)
                    self.add_measure(measure)
                    break
                else:
                    self.add_measure(measure)
                    current_measure_number += 1
                    remaining_duration -= measure.quarter_duration
                    if remaining_duration == 0:
                        break
            if barline_style:
                self.get_children_by_type(TreeMeasure)[-1].set_barline_style(barline_style)
            return current_measure_number

        if quarter_durations:
            if not hasattr(quarter_durations, '__iter__'):
                quarter_durations = [quarter_durations]
            elif isinstance(quarter_durations, str):
                raise TypeError()
            else:
                quarter_durations = quarter_durations

            current_measure_number = 1
            for duration in quarter_durations:
                current_measure_number = set_times(current_measure_number, duration)

            if list(times.keys()) != [] and current_measure_number < list(times.keys())[-1]:
                current_time = (4, 4)
                for key in times.keys():
                    if key <= current_measure_number:
                        current_time = times[key]

                for measure_number in range(current_measure_number, list(times.keys())[-1] + 1):
                    try:
                        current_time = times[measure_number]
                    except KeyError:
                        pass
                    self.add_measure(TreeMeasure(time=current_time))

        else:
            try:
                number_of_measures = list(times.keys())[-1]
                current_time = (4, 4)
                for measure_number in range(1, number_of_measures + 1):
                    try:
                        current_time = times[measure_number]
                    except KeyError:
                        pass
                    self.add_measure(TreeMeasure(time=current_time))
            except IndexError:
                pass

    def get_measure(self, number):
        if number == 0:
            raise ValueError('number can be positiv or negative integer but not 0')
        if number < 0:
            number += 1
        return self.get_children_by_type(TreeMeasure)[number - 1]

    def add_chord(self, measure_number, part_number, chord):
        if not isinstance(chord, TreeChord):
            raise TypeError('add_note note must be of type TreeChord not {}'.format(type(chord)))

        measure = self.get_children_by_type(TreeMeasure)[measure_number - 1]
        part = measure.get_children_by_type(TreePart)[part_number - 1]
        part.add_chord(chord)
        return chord

    def add_title(self, text, page=None, font_size=None, default_x=None, default_y=None, justify=None, valign=None,
                  **kwargs):
        if not page:
            page = 1
        if not font_size:
            font_size = 24
        if not default_x:
            default_x = int(self.page_style.page_width.value / 2)
        if not default_y:
            default_y = self.page_style.page_height.value - 43
        if not justify:
            justify = 'center'
        if not valign:
            valign = 'top'

        c = self.add_child(Credit(page=page))
        c.add_child(CreditType('title'))
        self._title = c.add_child(
            CreditWords(text, default_x=default_x, default_y=default_y, font_size=font_size, justify=justify,
                        valign=valign, **kwargs))

    def add_text(self, text, page=None, font_size=None, default_x=None, default_y=None, justify=None, valign=None,
                 **kwargs):

        if not page:
            page = 1
        if not font_size:
            font_size = 12
        if not default_x:
            default_x = 50
        if not default_y:
            default_y = self.page_style.page_height.value - 143
        if not justify:
            justify = 'left'
        if not valign:
            valign = 'top'

        c = self.add_child(Credit(page=page))
        self._title = c.add_child(
            CreditWords(text, default_x=default_x, default_y=default_y, font_size=font_size, justify=justify,
                        valign=valign, **kwargs))

    def remove_title(self):
        title_credit = [c for c in self.get_children_by_type(Credit) if
                        c.get_children_by_type(CreditType)[0].value == 'title']
        try:
            self.remove_child(title_credit[0])
        except IndexError:
            pass

    def add_subtitle(self, text, page=None, font_size=None, default_x=None, default_y=None, justify=None, valign=None,
                     **kwargs):
        if not page:
            page = 1
        if not font_size:
            font_size = 18
        if not default_x:
            default_x = int(self.page_style.page_width.value / 2)
        if not default_y:
            default_y = self.page_style.page_height.value - 93
        if not justify:
            justify = 'center'
        if not valign:
            valign = 'top'

        c = self.add_child(Credit(page=page))
        c.add_child(CreditType('subtitle'))
        self._subtitle = c.add_child(
            CreditWords(text, default_x=default_x, default_y=default_y, font_size=font_size, justify=justify,
                        valign=valign, **kwargs))

    def remove_subtitle(self):
        subtitle_credit = [c for c in self.get_children_by_type(Credit) if
                           c.get_children_by_type(CreditType)[0].value == 'subtitle']
        try:
            self.remove_child(subtitle_credit[0])
        except IndexError:
            pass

    def add_composer(self, text, page=None, font_size=None, default_x=None, default_y=None, justify=None, valign=None,
                     **kwargs):
        if not page:
            page = 1
        if not font_size:
            font_size = 12
        if not default_x:
            default_x = int(self.page_style.page_width.value - 50)
        if not default_y:
            default_y = self.page_style.page_height.value - 143
        if not justify:
            justify = 'right'
        if not valign:
            valign = 'top'

        c = self.add_child(Credit(page=page))
        c.add_child(CreditType('composer'))

        self._composer = c.add_child(
            CreditWords(text, default_x=default_x, default_y=default_y, font_size=font_size, justify=justify,
                        valign=valign, **kwargs))

    def update_measures(self):
        measures = self.get_children_by_type(TreeMeasure)
        for index, measure in enumerate(measures):
            measure.number = str(index + 1)
            if measure.time.force_hide:
                measure.hide_time_signature()
            elif measure.time.force_show:
                measure.show_time_signature()
            elif index == 0:
                measure.show_time_signature()
            elif measure.time.values == measures[index - 1].time.values:
                measure.hide_time_signature()
            else:
                measure.show_time_signature()

    def fill_with_rest(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.fill_with_rest()

    def preliminary_adjoin_rests(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.preliminary_adjoin_rests()

    def add_beats(self, list_of_beats=None):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.add_beats(list_of_beats)

    def get_beats(self):
        output = []
        for measure in self.get_children_by_type(TreeMeasure):
            output.extend(measure.get_beats())
        return output

    def quantize(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.quantize()

    def split_not_notatable(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.split_not_notatable()

    def implement_flags(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.implement_flags()

    def adjoin_ties(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.adjoin_ties()

    def adjoin_rests(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.adjoin_rests()

    def update_tuplets(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.update_tuplets()

    def substitute_sextoles(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.substitute_sextoles()

    def update_types(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.update_types()

    def update_dots(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.update_dots()

    def group_beams(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.group_beams()

    def implement_flags_2(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.implement_flags_2()

    def implement_flags_3(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.implement_flags_3()

    def chord_to_notes(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.chords_to_notes()

    def update_divisions(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.update_divisions()

    def update_accidentals(self, mode='normal'):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.update_accidentals(mode=mode)

    def update_durations(self):
        for measure in self.get_children_by_type(TreeMeasure):
            for part in measure.get_children_by_type(TreePart):
                part.update_durations()

    def finish(self):
        if not self._finished:
            self.update_measures()
            self.fill_with_rest()
            self.preliminary_adjoin_rests()
            self.add_beats()
            self.quantize()
            # self.adjoin_rests_in_beat()
            self.split_not_notatable()
            self.implement_flags()
            self.adjoin_ties()
            self.adjoin_rests()
            self.update_tuplets()
            self.substitute_sextoles()
            self.implement_flags_2()
            self.update_types()
            self.update_dots()
            self.group_beams()
            self.implement_flags_3()
            self.chord_to_notes()
            self.update_divisions()
            self.update_accidentals(mode=self.accidental_mode)
            self.update_durations()
            self.close_dtd()
            # for measure in self.get_children_by_type(TreeMeasure):
            #     for part in measure.get_children_by_type(TreePart):
            #         part.finish()
            # self.close_dtd()
            self._finished = True

    def to_string(self):
        self.finish()
        xml = self._to_xml()
        return et.tounicode(xml, pretty_print=True)

    def to_partwise(self):
        if not self._finished:
            raise Exception('timewise score must be finished before converting.')

        output = partwise.Score()
        output.version = self.version
        partwise_parts = []

        def get_ids():
            return [partwise_part.id for partwise_part in partwise_parts]

        for child in self.get_children():

            if isinstance(child, timewise.Measure):
                timewise_measure = child

                for timewise_part in timewise_measure.get_children_by_type(timewise.Part):
                    if timewise_part.id not in get_ids():
                        new_part = partwise.Part(id=timewise_part.id)
                        partwise_parts.append(new_part)
                    partwise_part = [p for p in partwise_parts if p.id == timewise_part.id][0]
                    new_measure = partwise.Measure()
                    for attribute in new_measure._ATTRIBUTES:
                        try:
                            new_measure.set_attribute(str(attribute), timewise_measure.get_attribute(str(attribute)))
                        except KeyError:
                            pass
                    for musicdata in timewise_part.get_children():
                        new_measure.add_child(musicdata)

                    partwise_part.add_child(new_measure)

            else:
                output.add_child(child)
        for partwise_part in partwise_parts:
            output.add_child(partwise_part)

        output.close_dtd()
        return output

    def extend(self, score):
        if not score.get_score_parts():
            raise ValueError('score {} cannot be empty'.format(score))

        if not self.get_score_parts():
            for score_part in score.get_score_parts():
                new_score_part = TreeScorePart(id=score_part.id)
                self.add_score_part(new_score_part)
                for part in new_score_part._parts:
                    new_score_part.add_part(part)

        else:
            my_ids = [score_part.id for score_part in self.get_score_parts()]
            other_ids = [score_part.id for score_part in score.get_score_parts()]
            difference = set.symmetric_difference(set(my_ids), set(other_ids))
            if difference:
                raise ValueError('two scores must have score_parts with same ids. Difference is {}'.format(difference))

        for measure in score.get_children_by_type(TreeMeasure):
            self.add_child(measure)
            for part in measure.get_children_by_type(TreePart):
                score_part = self.get_score_part(part.id)
                score_part.add_part(part)

        # if not self._identifications_added:
        #     self._add_identifications()



from musicscore.basic_functions import replace_dash
from musicscore.musicxml.types.simple_type import *


class AttributeAbstract(object):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ATTRIBUTES = []

    def generate_attribute(self, attribute_name, attribute_value, type_string=None):
        property_name = replace_dash(attribute_name)

        exec('def getter(self): return self.get_attribute("{}")'.format(attribute_name))
        exec('''def setter(self, value): 
    if value is None:
        self.remove_attribute('{}')
    else:
        if {} is not None:
            {}(value)
        self.set_attribute('{}', value)'''.format(attribute_name, type_string, type_string, attribute_name))
        exec('AttributeAbstract.{} = property(getter, setter)'.format(property_name))
        self._ATTRIBUTES.insert(0, attribute_name)
        if isinstance(attribute_value, str):
            exec('self.{} = "{}"'.format(property_name, attribute_value))
        else:
            exec('self.{} = {}'.format(property_name, attribute_value))

        return property_name



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Underline(AttributeAbstract):
    """"""

    def __init__(self, underline=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('underline', underline, "TypeNumberOfLines")


class Overline(AttributeAbstract):
    """"""

    def __init__(self, overline=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('overline', overline, "TypeNumberOfLines")


class LineThrough(AttributeAbstract):
    """"""

    def __init__(self, line_through=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('line-through', line_through, "TypeNumberOfLines")


class TextDecoration(Underline, Overline, LineThrough):
    """
    The text-decoration attribute group is based on the similar feature in XHTML and CSS. It allows for text to be
    underlined, overlined, or struck-through. It extends the CSS version by allow double or triple lines instead of just
    being on or off.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Version(AttributeAbstract):
    def __init__(self, version='1.0', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('version', version, 'Token')


class DocumentAttributes(Version):
    """
    The document-attributes attribute group is used to specify the attributes for an entire MusicXML document. Currently
    this is used for the version attribute.

    The version attribute was added in Version 1.1 for the score-partwise and score-timewise documents. It provides an
    easier way to get version information than through the MusicXML public ID. The default value is 1.0 to make it
    possible for programs that handle later versions to distinguish earlier version files reliably. Programs that write
    MusicXML 1.1 or later files should set this attribute.
    """
    def __init__(self, version='1.0', *args, **kwargs):
        super().__init__(version=version, *args, **kwargs)

# class FontSize(AttributeAbstract):
#     """"""
#
#     def __init__(self, font_size=None, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.generate_attribute('font-size', font_size, 'FontSizeType')



from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.attributes.trillsound import TrillSound
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeYesNo


class ComplexTypeHorizontalTurn(ComplexType, PrintStyle, Placement, TrillSound):
    """
    The horizontal-turn type represents turn elements that are horizontal rather than vertical. These are empty elements
    with print-style, placement, trill-sound, and slash attributes. If the slash attribute is yes, then a vertical line
    is used to slash the turn; it is no by default.
    """

    def __init__(self, tag, slash='no', *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.slash = slash

    @property
    def slash(self):
        return self.get_attribute('slash')

    @slash.setter
    def slash(self, value):
        if value is None:
            self.remove_attribute('slash')
        else:
            TypeYesNo(value)
            self._ATTRIBUTES.insert(0, 'slash')
            self.set_attribute('slash', value)



from musicscore.musicxml.attributes.dahsedformatting import DashedFormatting
from musicscore.musicxml.attributes.linelength import LineLength
from musicscore.musicxml.attributes.lineshape import LineShape
from musicscore.musicxml.attributes.linetype import LineType
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.types.complextypes.complextype import Empty


class ComplexTypeEmptyLine(Empty, LineShape, LineType, LineLength, DashedFormatting, PrintStyle, Placement):
    """The empty-line type represents an empty element with line-shape, line-type, line-length, dashed-formatting,
    print-style and placement attributes."""

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.attributes.bezier import Bezier
from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.dahsedformatting import DashedFormatting
from musicscore.musicxml.attributes.linetype import LineType
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.orientation import Orientation
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.types.complextypes.complextype import Empty
from musicscore.musicxml.types.simple_type import TypeStartStopContinue, TypeNumberLevel


class ComplexTypeSlur(Empty, LineType, DashedFormatting, Position, Placement, Orientation, Bezier, Color,
                      OptionalUniqueId):
    """
    Slur types are empty. Most slurs are represented with two elements: one with a start type, and one with a stop type.
    Slurs can add more elements using a continue type. This is typically used to specify the formatting of cross-system
    slurs, or to specify the shape of very complex slurs.
    """

    def __init__(self, tag, type, number=1, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.type = type
        self.number = number

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeStartStopContinue(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)

    @property
    def number(self):
        return self.get_attribute('number')

    @number.setter
    def number(self, value):
        if value is None:
            self.remove_attribute('number')
        else:
            TypeNumberLevel(value)
            self._ATTRIBUTES.insert(0, 'number')
            self.set_attribute('number', value)



from musicscore.musicxml.attributes.elementposition import ElementPosition
from musicscore.musicxml.attributes.linkattributes import LinkAttributes
from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import Token


class ComplexTypeLink(ComplexType, LinkAttributes, ElementPosition, Position):
    """The link type serves as an outgoing simple XLink. It is also used to connect a MusicXML score with a MusicXML
    opus. If a relative link is used within a document that is part of a compressed MusicXML file, the link is relative
    to the  root folder of the zip file."""

    def __init__(self, tag, name=None, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.name = name

    @property
    def name(self):
        return self.get_attribute('name')

    @name.setter
    def name(self, value):
        if value is None:
            self.remove_attribute('name')
        else:
            Token(value)
            self._ATTRIBUTES.insert(0, 'name')
            self.set_attribute('name', value)



from musicscore.musicxml.elements.note import Notehead, TimeModification, Stem, Notations, Type
from musicscore.musicxml.groups.common import Voice
from musicscore.musicxml.types.complextypes.timemodification import ActualNotes, NormalNotes


class TreeChordFlag2(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class FingerTremoloFlag2(TreeChordFlag2):
    def __init__(self, tremolo_chord, number=3, mode='conventional', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mode = None
        self.mode = mode
        self._tremolo_chord = None
        self.tremolo_chord = tremolo_chord
        self.number = number
        self.slur = None

    def __deepcopy__(self, memodict={}):
        return self.__class__(tremolo_chord=self.tremolo_chord.tremolo_flag_copy(), number=self.number, mode=self.mode)

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, val):
        permitted = ['conventional', 'modern']
        if val not in permitted:
            raise ValueError('mode.value {} must be of in {}'.format(type(val), permitted))
        self._mode = val

    @property
    def tremolo_chord(self):
        return self._tremolo_chord

    @tremolo_chord.setter
    def tremolo_chord(self, val):
        if val is not None and self.mode == 'modern':
            val.is_finger_tremolo = True
            val.add_child(Stem('none'))
            val.set_manual_type('quarter', size='full')
            tm = TimeModification()
            tm.add_child(ActualNotes(0))
            tm.add_child(NormalNotes(1))
            val.add_child(tm)
            val.is_adjoinable = False
            val.quarter_duration = 0

        self._tremolo_chord = val

    def _implement_conventional(self, chord):

        output = [chord]
        chord.remove_tie('start')
        chord.remove_tie('stop')

        output[0].quarter_duration /= 2
        output[0].add_tremolo(type='start')

        if not self.tremolo_chord.tremoli:
            self.tremolo_chord.add_tremolo(type='stop')

        self.tremolo_chord.quarter_duration = chord.quarter_duration

        # self.tremolo_chord.parent_voice = chord.parent_voice
        self.tremolo_chord.parent_beat = chord.parent_beat
        # v = chord.get_children_by_type(Voice)[0]
        #
        # if not self.tremolo_chord.get_children_by_type(Voice):
        #     self.tremolo_chord.add_child(v)

        output.insert(1, self.tremolo_chord)

        return output

    def _implement_modern(self, chord):

        if self.tremolo_chord.midis[0].value < chord.midis[0].value:
            chord.add_words('\uF415', font_family='bravura', font_size=16, relative_x=30, relative_y=-50)
            chord.set_tie_orientation('over')
        else:
            chord.add_words('\uF417', font_family='bravura', font_size=16, relative_x=30, relative_y=-50)
            chord.set_tie_orientation('under')
        self.tremolo_chord.parent_voice = chord.parent_voice
        self.tremolo_chord.parent_beat = chord.parent_beat
        v = chord.get_children_by_type(Voice)[0]
        if not self.tremolo_chord.get_children_by_type(Voice):
            self.tremolo_chord.add_child(v)

        return [chord, self.tremolo_chord]

    def implement(self, chord):
        if self.mode == 'modern':
            return self._implement_modern(chord)
        elif self.mode == 'conventional':
            return self._implement_conventional(chord)





from unittest import TestCase
from musicscore.musicxml.elements.fullnote import DisplayStep, DisplayOctave, Rest


class TestRest(TestCase):
    def setUp(self):
        self.rest = Rest()

    def test_rest(self):
        result = '''<rest/>
'''

        self.assertEqual(self.rest.to_string(), result)

    def test_rest_2(self):
        self.rest.add_child(DisplayStep('C'))
        self.rest.add_child(DisplayOctave(4))
        result = '''<rest>
  <display-step>C</display-step>
  <display-octave>4</display-octave>
</rest>
'''
        self.assertEqual(self.rest.to_string(), result)

    def test_rest_3(self):
        self.rest.add_child(DisplayStep('C'))
        self.rest.display_step.value = 'D'
        self.rest.add_child(DisplayOctave(7))
        result = '''<rest>
  <display-step>D</display-step>
  <display-octave>7</display-octave>
</rest>
'''
        self.assertEqual(self.rest.to_string(), result)




from musicscore.basic_functions import substitute
from unittest import TestCase


class Test(TestCase):
    def test_substitute(self):
        input_list = [1, 2, 3, 4, 5]
        new_elements = ['a', 'b', 'c']
        result = [1, 2, 'a', 'b', 'c', 4, 5]

        self.assertEqual(substitute(input_list, 3, new_elements), result)

        result = [1, 2, 'z', 4, 5]
        new_elements = 'z'
        self.assertEqual(substitute(input_list, 3, new_elements), result)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract, TypeLineEnd
from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.dahsedformatting import DashedFormatting
from musicscore.musicxml.attributes.linetype import LineType
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeNumberLevel, TypeStartStopContinue


class LineEnd(AttributeAbstract):
    """"""

    def __init__(self, line_end, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('line-end', line_end, 'TypeLineEnd')
        TypeLineEnd


class EndLength(AttributeAbstract):
    """"""

    def __init__(self, end_length=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('end-length', end_length, 'TypeTenth')


class ComplexTypeBracket(ComplexType, LineEnd, EndLength, LineType, DashedFormatting, Position, Color,
                         OptionalUniqueId):

    def __init__(self, tag, type, number=1, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.type = type
        self.number = number

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeStartStopContinue(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)

    @property
    def number(self):
        return self.get_attribute('number')

    @number.setter
    def number(self, value):
        if value is None:
            self.remove_attribute('number')
        else:
            TypeNumberLevel(value)
            self._ATTRIBUTES.insert(0, 'number')
            self.set_attribute('number', value)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def make_t(self, *args):
        sf = SimpleFormat(quarter_durations=[1, 1, 1, 1], midis=[0, 0, 0, 0])
        for i in range(3):
            sf.chords[i].is_adjoinable = args[i]
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)
        self.score.add_subtitle(str([chord.is_adjoinable for chord in sf.chords]))

    def test_1(self):
        self.make_t(False, False, False)
        result_path = path + '_test_1'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_2(self):
        self.make_t(False, False, True)
        result_path = path + '_test_2'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_3(self):
        self.make_t(False, True, False)
        result_path = path + '_test_3'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_4(self):
        self.make_t(False, True, True)
        result_path = path + '_test_4'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_5(self):
        self.make_t(True, False, False)
        result_path = path + '_test_5'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_6(self):
        self.make_t(True, False, True)
        result_path = path + '_test_6'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_7(self):
        self.make_t(True, True, False)
        result_path = path + '_test_7'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_8(self):
        self.make_t(True, True, True)
        result_path = path + '_test_8'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_9(self):
        sf = SimpleFormat(quarter_durations=[0.2, 0.8, 1, 1], midis=[0, 0, 0, 60])
        sf.to_stream_voice().add_to_score(self.score)
        result_path = path + '_test_9'
        self.score.write(path=result_path)
        # TestScore().assert_template(result_path=result_path)

    def test_10(self):
        sf = SimpleFormat(quarter_durations=[0.5, 2.5], midis=[60, 0])
        self.score.set_time_signatures(quarter_durations=[3])
        sf.to_stream_voice().add_to_score(self.score)
        result_path = path + '_test_10'
        self.score.write(path=result_path)



from unittest import TestCase

from musicscore.dtd.dtd import DTDError, Sequence, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from lxml import etree as et


class TestXMLElement(TestCase):

    def setUp(self):
        self.xml = XMLElement('root')

    def test_tag(self):
        self.assertEqual(self.xml.tag, 'root')

    def test_text(self):
        self.xml.text = 2
        self.assertEqual(self.xml.text, 2)

    def test_add_child(self):
        child = 'child'
        with self.assertRaises(DTDError):
            self.xml.add_child(child)
        #
        # child = XMLElement('child')
        # self.xml.add_child(child)
        # self.assertEqual(self.xml.get_children(), [child])

    # def test_remove_old_child_by_tag(self):
    #     children = [XMLElement('child-1'), XMLElement('child-2')]
    #     for child in children:
    #         self.xml.add_child(child)
    #     self.xml.remove_old_child_by_tag(children[0].tag)
    #     self.assertEqual(self.xml.get_children(), [children[1]])

    # def test_to_xml(self):
    #     children = [XMLElement('child-1'), XMLElement('child-2')]
    #     for child in children:
    #         self.xml.add_child(child)
    #     xml = self.xml._to_xml()
    #     self.assertIsInstance(xml, et._Element)

    # def test_to_string(self):
    #     class Child1(XMLElement):
    #         def __init__(self):
    #             super().__init__('child-1')
    #
    #     class Child2(XMLElementGroup):
    #         def __init__(self):
    #             super().__init__('child-2')
    #
    #     class Sibling(XMLElement):
    #         def __init__(self):
    #             super().__init__('child-2')
    #
    #     children = [Child1(), Child2()]
    #     children[1].add_sibling(Sibling())
    #     children[1].add_sibling(Sibling())
    #
    #     for child in children:
    #         self.xml.add_child(child)
    #     self.assertEqual(self.xml.to_string(), '<root>\n  <child-1/>\n  <child-2/>\n  <child-2/>\n</root>\n')
    #
    # def test_text_to_string(self):
    #     self.xml.text = 'test'
    #     self.assertEqual(self.xml.to_string(), '<root>test</root>\n')

    # def test_find_child_by_tag(self):
    #     children = [XMLElement('child-1'), XMLElement('child-2'), XMLElement('child-3')]
    #     for child in children:
    #         self.xml.add_child(child)
    #     self.assertEqual(self.xml.find_child_by_tag('child-3').tag, 'child-3')

    # def test_sort_children(self):
    #
    #     class Child1(XMLElement):
    #         def __init__(self):
    #             super().__init__('child1')
    #
    #     class Child2(XMLElement):
    #         def __init__(self):
    #             super().__init__('child2')
    #
    #     class Child3(XMLElementGroup):
    #         def __init__(self):
    #             super().__init__('child3')
    #
    #     children = [Child1(), Child2(), Child3(), Child2()]
    #     for child in children:
    #         self.xml.add_child(child)
    #
    #     self.xml._CHILDREN_ORDERED = True
    #     self.xml._CHILDREN_TYPES = [Child2, Child3, Child1]
    #
    #     self.xml._sort_children()
    #     self.assertEqual([child.tag for child in self.xml.get_children()], ['child2', 'child3', 'child1', 'child2'])

    # def test_attributes(self):
    #     self.xml.set_attribute('att1', 2)
    #     self.xml.set_attribute('att2', 'a')
    #     self.xml.set_attribute('att3', None)
    #     result = '<root att1="2" att2="a" att3="None"/>\n'
    #     self.assertEqual(self.xml.to_string(), result)



from unittest import TestCase

from musicscore.musicxml.elements.fullnote import Pitch
from musicscore.musicxml.elements.note import Note, Duration, Notations
from musicscore.musicxml.types.complextypes.articulations import Accent, Staccato
from musicscore.musicxml.types.complextypes.notations import Articulations


class Test(TestCase):
    def setUp(self):
        self.note = Note()
        self.note.add_child(Duration())
        self.note.add_child(Pitch())

    def test_1(self):
        notations = self.note.add_child(Notations())
        articulations = notations.add_child(Articulations())
        articulations.add_child(Accent())
        articulations.add_child(Staccato(relative_y=20))

        result = """<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>1</duration>
  <notations>
    <articulations>
      <accent/>
      <staccato relative-y="20"/>
    </articulations>
  </notations>
</note>
"""
        self.assertEqual(self.note.to_string(), result)



from unittest import TestCase

from musicscore.musicxml.attributes.document_attributes import Version
from musicscore.musicxml.elements.xml_element import XMLElement


class TestTimewise(TestCase):
    def setUp(self):
        class Bla(XMLElement, Version):
            def __init__(self, *args, **kwargs):
                super().__init__(tag='bal', *args, **kwargs)
        self.bla = Bla()

    def test_version(self):

        self.bla.version = '10'
        result = '''<bal version="10"/>
'''
        self.assertEqual(self.bla.to_string(), result)



from unittest import TestCase
import random
import os

from musicscore.musicstream.streamvoice import SimpleFormat, TreeMeasure, TreePart
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        random.seed(1)
        durations = [random.random() + random.random() for i in range(10)]
        self.sf = SimpleFormat(quarter_durations=durations)
        self.score = TreeScoreTimewise()

    def test_1(self):
        xml_path = path + '_test_1.xml'
        self.sf.to_stream_voice().add_to_score(self.score, part_number=1)
        self.sf.to_stream_voice().add_to_score(self.score, part_number=2)
        for measure in self.score.get_children_by_type(TreeMeasure):
            measure.get_children_by_type(TreePart)[-1].forbidden_divisions = [8]
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_2(self):
        xml_path = path + '_test_2.xml'
        self.score.max_division = 6
        self.sf.to_stream_voice().add_to_score(self.score, part_number=1)
        self.sf.to_stream_voice().add_to_score(self.score, part_number=2)
        for measure in self.score.get_children_by_type(TreeMeasure):
            measure.get_children_by_type(TreePart)[-1].forbidden_divisions = [5]
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)

    def test_3(self):
        xml_path = path + '_test_3.xml'
        self.score.max_division = 6
        self.score.forbidden_divisions = [8]
        self.sf.to_stream_voice().add_to_score(self.score, part_number=1)
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)



"""
    <xs:complexType name="encoding">
        <xs:choice minOccurs="0" maxOccurs="unbounded">
            <xs:element name="software" type="xs:string"/>
            <xs:element name="encoding-description" type="xs:string"/>
            <xs:element name="supports" type="supports"/>
        </xs:choice>
    </xs:complexType>
"""
from musicscore.dtd.dtd import Choice, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.supports import ComplexTypeSupports


class EncodingDate(XMLElement):
    """
    <xs:element name="encoding-date" type="yyyy-mm-dd"/>
    """
    _TAG = 'encoding-date'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        raise NotImplementedError(self._TAG)


class Encoder(object):
    """
    <xs:element name="encoder" type="typed-text"/>
    """
    _TAG = 'encoder'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        raise NotImplementedError(self._TAG)


class Software(object):
    """
    <xs:element name="software" type="xs:string"/>
    """
    _TAG = 'software'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        raise NotImplementedError(self._TAG)


class EncodingDescription(object):
    """
    <xs:element name="encoding-description" type="xs:string"/>
    """
    _TAG = 'encoding-description'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        raise NotImplementedError(self._TAG)


class Supports(ComplexTypeSupports):
    """
    <xs:element name="supports" type="supports"/>
    """
    _TAG = 'supports'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypeEncoding(ComplexType):
    """
    The encoding element contains information about who did the digital encoding, when, with what software, and in what
     aspects. Standard type values for the encoder element are music, words, and arrangement, but other types may be
     used. The type attribute is only needed when there are multiple encoder element.
    """
    _DTD = Choice(
        Element(EncodingDate),
        Element(Encoder),
        Element(Software),
        Element(EncodingDescription),
        Element(Supports),
        min_occurrence=0,
        max_occurrence=None,
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.dtd.dtd import GroupReference, Sequence
from musicscore.musicxml.groups.margins import AllMargins
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeMarginType


class ComplexTypePageMargins(ComplexType):
    """Page margins are specified either for both even and odd pages, or via separate odd and even page number values.
    The type attribute is not needed when used as part of a print element. If omitted when the page-margins type is
    used in the defaults element, "both" is the default value."""

    _DTD = Sequence(
        GroupReference(AllMargins)
    )

    def __init__(self, tag, type_=None, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.type = type_

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeMarginType(value)
            self._ATTRIBUTES = ['type']
            self.set_attribute('type', value)



from musicscore.dtd.dtd import Choice, Sequence, GroupReference, Element
from musicscore.musicxml.attributes.font import Font
from musicscore.musicxml.attributes.justify import Justify
from musicscore.musicxml.attributes.leveldisplay import Parentheses
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.printstyle import PrintStyleAlign
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType, Empty
from musicscore.musicxml.types.complextypes.timemodification import ComplexTypeTimeModification
from musicscore.musicxml.types.complextypes.tuplet import Bracket, ShowNumber
from musicscore.musicxml.types.simple_type import String, TypeNoteTypeValue, TypeBeamValue, TypeBeamLevel, TypeStartStop


class PerMinute(ComplexType, String, Font):
    """The per-minute type can be a number, or a text description including numbers. If a font is specified, it
    overrides the font specified for the overall metronome element. This allows separate specification of a music font
    for the beat-unit and a text font for the numeric value, in cases where a single metronome font is not used."""

    _TAG = 'per-minute'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class BeatUnit(XMLElement, TypeNoteTypeValue):
    """The beat-unit element indicates the graphical note type to use in a metronome mark."""

    _TAG = 'beat-unit'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class BeatUnitDot(Empty):
    """The beat-unit-dot element is used to specify any augmentation dots for a metronome mark note.."""

    _TAG = 'beat-unit'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


"""
The beat-unit group combines elements used repeatedly in the metronome element to specify a note within a metronome 
mark.
"""
GroupBeatUnit = Sequence(
    Element(BeatUnit),
    Element(BeatUnitDot, min_occurrence=0, max_occurrence=None)
)


class BeatUnitTied(ComplexType):
    """The beat-unit-tied type indicates a beat-unit within a metronome mark that is tied to the preceding beat-unit.
    This allows or two or more tied notes to be associated with a per-minute value in a metronome mark, whereas the
    metronome-tied element is restricted to metric relationship marks."""
    _TAG = 'beat-unit-tied'
    _DTD = GroupBeatUnit

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class MetronomeArrows(Empty):
    """If the metronome-arrows element is present, it indicates that metric modulation arrows are displayed on both
    sides of the metronome mark."""

    _TAG = 'metronome-arrows'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class MetronomeType(XMLElement, TypeNoteTypeValue):
    """The metronome-type element works like the type element in defining metric relationships."""

    _TAG = 'metronome-type'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class MetronomeDot(Empty):
    """The metronome-dot element works like the dot element in defining metric relationships."""

    _TAG = 'metronome-dot'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class MetronomeBeam(ComplexType, TypeBeamValue):
    """The metronome-beam type works like the beam type in defining metric relationships, but does not include all the
    attributes available in the beam type."""

    _TAG = 'metronome-beam'

    def __init__(self, value=None, number=1, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)
        self.number = number

    @property
    def number(self):
        return self.get_attribute('number')

    @number.setter
    def number(self, value):
        if value is None:
            self.remove_attribute('number')
        else:
            TypeBeamLevel(value)
            self._ATTRIBUTES.insert(0, 'number')
            self.set_attribute('number', value)


class MetronomeTied(ComplexType):
    """The metronome-tied indicates the presence of a tie within a metric relationship mark. As with the tied element,
    both the start and stop of the tie should be specified, in this case within separate metronome-note elements."""

    _TAG = 'metronome-tied'

    def __init__(self, type_='start', *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        self.type = type_

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeStartStop(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)


class MetronomeTuplet(ComplexTypeTimeModification, Bracket, ShowNumber):
    """The metronome-tuplet type uses the same element structure as the time-modification element along with some
    attributes from the tuplet element."""

    _TAG = 'metronome-tuplet'

    def __init__(self, type_='start', *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        self.type = type_

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeStartStop(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)


class MetronomeNote(ComplexType):
    """The metronome-note type defines the appearance of a note within a metric relationship mark."""
    _DTD = Sequence(
        Element(MetronomeType),
        Element(MetronomeDot, min_occurrence=0, max_occurrence=None),
        Element(MetronomeBeam, min_occurrence=0, max_occurrence=None),
        Element(MetronomeTied, min_occurrence=0),
        Element(MetronomeTuplet, min_occurrence=0)
    )
    _TAG = 'metronome-note'

    def __init__(self, args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class MetronomeRelation(XMLElement, String):
    """The metronome-relation element describes the relationship symbol that goes between the two sets of
    metronome-note elements. The currently allowed value is equals, but this may expand in future versions. If the
    element is empty, the equals value is used."""

    _TAG = 'metronome-relation'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class ComplexTypeMetronome(ComplexType, PrintStyleAlign, Justify, Parentheses, OptionalUniqueId):
    """
    The metronome type represents metronome marks and other metric relationships. The beat-unit group and per-minute
    element specify regular metronome marks. The metronome-note and metronome-relation elements allow for the
    specification of metric modulations and other metric relationships, such as swing tempo marks where two eighths are
    equated to a quarter note / eighth note triplet. Tied notes can be represented in both types of metronome marks by
    using the beat-unit-tied and metronome-tied elements. The parentheses attribute indicates whether or not to put
    the metronome mark in parentheses; its value is no if not specified.
    """
    _DTD = Choice(
        Sequence(
            GroupReference(GroupBeatUnit),
            Element(BeatUnitTied, min_occurrence=0, max_occurrence=None),
            Choice(
                Element(PerMinute),
                Sequence(
                    GroupReference(GroupBeatUnit),
                    Element(BeatUnitTied, min_occurrence=0, max_occurrence=None),
                )
            )
        ),
        Sequence(
            Element(MetronomeArrows, min_occurrence=0),
            Element(MetronomeNote, max_occurrence=None),
            Sequence(
                Element(MetronomeRelation),
                Element(MetronomeNote),
                min_occurrence=0
            )
        )
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Color(AttributeAbstract):
    """
    The color type indicates the color of an element. Color may be represented as hexadecimal RGB triples, as in HTML,
    or as hexadecimal ARGB tuples, with the A indicating alpha of transparency. An alpha value of 00 is totally
    transparent; FF is totally opaque. If RGB is used, the A value is assumed to be FF.  For instance, the RGB value
    "#800080" represents purple. An ARGB value of "#40800080" would be a transparent purple.  As in SVG 1.1, colors are
    defined in terms of the sRGB color space (IEC 61966).
    """

    def __init__(self, color=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('color', color, 'TypeColor')



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Placement(AttributeAbstract):
    def __init__(self, placement=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('placement', placement, 'TypeAboveBelow')



from unittest import TestCase

from musicscore.musicxml.groups.musicdata import Direction
from musicscore.musicxml.types.complextypes.direction import DirectionType
from musicscore.musicxml.types.complextypes.directiontype import Wedge


class Test(TestCase):
    def test_1(self):
        wedge_object = Wedge('crescendo')
        direction = Direction(placement='below')
        direction_type = direction.add_child(DirectionType())
        direction_type.add_child(wedge_object)

        result = """<direction placement="below">
  <direction-type>
    <wedge type="crescendo"/>
  </direction-type>
</direction>
"""
        self.assertEqual(direction.to_string(), result)



import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = str(os.path.abspath(__file__).split('.')[0])
xml_path = path + '.xml'
score = TreeScoreTimewise()
# quarter_durations = [Fraction(918, 515), Fraction(382, 485), Fraction(658, 429), Fraction(577, 449), Fraction(471, 874),
#                      Fraction(998, 963), Fraction(589, 759), Fraction(1096, 669), Fraction(589, 414),
#                      Fraction(589, 594), Fraction(169, 140), Fraction(944, 941), Fraction(1431, 923),
#                      Fraction(658, 555), Fraction(658, 481), Fraction(1105, 784), Fraction(221, 392),
#                      Fraction(221, 196), Fraction(221, 112), Fraction(663, 784), Fraction(663, 392), Fraction(221, 784)]
#
# limit = 11
# quarter_durations = [sum(quarter_durations[:-limit])] + quarter_durations[-limit:]
# print([float(qd) for qd in quarter_durations])
quarter_durations = [0.99999479229065, 1.0031880977683316, 1.5503791982665223, 1.1855855855855857, 1.367983367983368,
                     1.409438775510204, 0.5637755102040817, 1.1275510204081634, 1.9732142857142858, 0.8456632653061225,
                     1.691326530612245, 0.28188775510204084]
# quarter_durations = [4, 0]
# sf = SimpleFormat(durations=quarter_durations, midis=[60, 0])
sf = SimpleFormat(quarter_durations=quarter_durations)
score.set_time_signatures(round(sum(quarter_durations)))
# print([float(chord.quarter_duration) for chord in sf.chords])
sf.to_stream_voice().add_to_score(score)
#
# # score.update_measures()
# # score.fill_with_rest()
# # # score.preliminary_adjoin_rests()
# # # score.add_beats()
# # # score.quantize()
# print(score.get_children_by_type(TreeMeasure))
# print([float(chord.quarter_duration) for chord in score.get_measure(-1).get_part(1).chords])
score.write(xml_path)
TreeChord()



# from unittest import TestCase
# from musicscore.partwise import PartPartwise
#
#
# class TestGeneratePartId(TestCase):
#
#     def test_generate_part_id(self):
#         PartPartwise.reset_ids()
#
#         def get_id(number):
#             for i in range(number):
#                 part = PartPartwise()
#         get_id(1)
#         PartPartwise.reset_ids()
#         get_id(3)
#         part = PartPartwise(id='p10')
#         get_id(1)
#         result = ['p1', 'p2', 'p3', 'p10', 'p4']
#         self.assertEqual(part._ids, result)
#         with self.assertRaises(ValueError):
#             part = PartPartwise(id='p4')



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.sf = SimpleFormat([1, 2, 3, 4, 3, 2, 1])
        self.score = TreeScoreTimewise()

    def test_1(self):
        self.sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + 'test_1.xml'
        self.score.write(xml_path)
        TestScore().assert_template(result_path=xml_path)

    def test_2(self):
        expected = 16
        actual = self.sf.quarter_duration
        self.assertEqual(expected, actual)



'''
    <xs:complexType name="identification">
        <xs:sequence>
            <xs:element name="creator" type="typed-text" minOccurs="0" maxOccurs="unbounded">
            </xs:element>
            <xs:element name="rights" type="typed-text" minOccurs="0" maxOccurs="unbounded">
            </xs:element>
            <xs:element name="encoding" type="encoding" minOccurs="0"/>
            <xs:element name="source" type="xs:string" minOccurs="0">
            </xs:element>
            <xs:element name="relation" type="typed-text" minOccurs="0" maxOccurs="unbounded">
                <xs:annotation>
                    <xs:documentation>
                    </xs:documentation>
                </xs:annotation>
            </xs:element>
            <xs:element name="miscellaneous" type="miscellaneous" minOccurs="0"/>
        </xs:sequence>
    </xs:complexType>
'''
from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.encoding import ComplexTypeEncoding
from musicscore.musicxml.types.complextypes.typedtext import ComplexTypeTypedText
from musicscore.musicxml.types.simple_type import String


class Rights(ComplexTypeTypedText):
    """The rights element is borrowed from Dublin Core. It contains copyright and other intellectual property notices.
    Words, music, and derivatives can have different types, so multiple rights tags with different type attributes are
    supported. Standard type values are music, words, and arrangement, but other types may be used. The type attribute
    is only needed when there are multiple rights elements."""
    _TAG = 'rights'

    def __init__(self, type=None, *args, **kwargs):
        super().__init__(tag=self._TAG, type=type, *args, **kwargs)


class Creator(ComplexTypeTypedText):
    """
    The creator element is borrowed from Dublin Core. It is used for the creators of the score. The type attribute is
    used to distinguish different creative contributions. Thus, there can be multiple creators within an identification.
    Standard type values are composer, lyricist, and arranger. Other type values may be used for different types of
    creative roles. The type attribute should usually be used even if there is just a single creator element. The
    MusicXML format does not use the creator / contributor distinction from Dublin Core.
    """
    _TAG = 'creator'

    def __init__(self, type=None, *args, **kwargs):
        super().__init__(tag=self._TAG, type=type, *args, **kwargs)


class Relation(ComplexTypeTypedText):
    """
    A related resource for the music that is encoded. This is similar to the Dublin Core relation element.
    Standard type values are music, words, and arrangement, but other types may be used.
    """
    _TAG = 'relation'

    def __init__(self, type=None, *args, **kwargs):
        super().__init__(tag=self._TAG, type=type, *args, **kwargs)


class Encoding(ComplexTypeEncoding):
    _TAG = 'encoding'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Source(XMLElement, String):
    """
    The source for the music that is encoded. This is similar to the Dublin Core source element.
    """
    _TAG = 'source'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Miscellaneus(XMLElement):
    """
    <xs:element name="miscellaneus" type="miscellaneus"/>
    """
    _TAG = 'miscellaneus'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)
        raise NotImplementedError(self._TAG)


class ComplexTypeIdentification(ComplexType):
    """
    Identification contains basic metadata about the score. It includes the information in MuseData headers
    that may apply at a score-wide, movement-wide, or part-wide level. The creator, rights,
    source, and relation elements are based on Dublin Core.
    """
    _DTD = Sequence(
        Element(Creator,
                min_occurrence=0,
                max_occurrence=None),
        Element(Rights,
                min_occurrence=0,
                max_occurrence=None),
        Element(Encoding,
                min_occurrence=0),
        Element(Source,
                min_occurrence=0),
        Element(Relation,
                min_occurrence=0,
                max_occurrence=None),
        Element(Miscellaneus,
                min_occurrence=0)
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



import uuid

from musicscore.musictree.midi import G, D, A, E, C, MidiNote, Midi, B, F, midi_to_frequency, frequency_to_midi
from musicscore.musictree.treeclef import TreeClef, TREBLE_CLEF, ALTO_CLEF, BASS_CLEF
from musicscore.musicxml.types.complextypes.midiinstrument import ComplexTypeMidiInstrument
from musicscore.musicxml.types.complextypes.scorepart import PartName, PartAbbreviation


class TreeInstrument(ComplexTypeMidiInstrument):
    _TAG = 'midi-instrument'

    def __init__(self, name, abbreviation=None, number=None, *args, **kwargs):
        super().__init__(tag=self._TAG, id_='inst' + str(uuid.uuid4()), *args, **kwargs)
        self._part_name = PartName(name=name)
        self._part_abbreviation = PartAbbreviation()
        self._standard_clef = None
        self._number = None
        self.abbreviation = abbreviation
        self.number = number

    @property
    def part_name(self):
        return self._part_name

    @property
    def part_abbreviation(self):
        return self._part_abbreviation

    @property
    def number(self):
        return self._number

    @number.setter
    def number(self, val):
        if self._number is not None:
            raise AttributeError('number can only be set once')

        if val is not None and not isinstance(val, int):
            raise TypeError('number.value must be of type int not{}'.format(type(val)))
        self._number = val

        if self._number is not None:
            self.name += ' ' + str(self._number)
            self.abbreviation += ' ' + str(self._number)

    @property
    def name(self):
        return self._part_name.name

    @name.setter
    def name(self, val):
        self._part_name.name = val

    @property
    def abbreviation(self):
        return self._part_abbreviation.value

    @abbreviation.setter
    def abbreviation(self, val):
        self._part_abbreviation.value = val

    @property
    def standard_clef(self):
        return self._standard_clef

    @standard_clef.setter
    def standard_clef(self, val):
        if val and not isinstance(val, TreeClef):
            raise TypeError('standard_clef.value must be of type TreeClef not{}'.format(type(val)))
        self._standard_clef = val


class String(object):
    def __init__(self, number, tuning, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tuning = None
        self.number = number
        self.tuning = tuning

    @property
    def tuning(self):
        return self._tuning

    @tuning.setter
    def tuning(self, val):
        if not isinstance(val, Midi):
            raise TypeError('tuning.value must be of type Midi not{}'.format(type(val)))
        self._tuning = val

    def get_step(self, number):
        step = self.tuning.__deepcopy__()
        step.transpose(number)
        return step


class NaturalInstrument(TreeInstrument):

    def __init__(self, key, a4=440, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._a4 = None
        self._key = None
        self._transposition = None
        self.a4 = a4
        self.key = key

    @property
    def a4(self):
        return self._a4

    @a4.setter
    def a4(self, val):
        try:
            float(val)
        except AttributeError:
            raise TypeError()
        self._a4 = val

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, val):
        if not isinstance(val, MidiNote):
            raise TypeError('key.value must be of type MidiNote not{}'.format(type(val)))
        self._key = val

    @property
    def transposition(self):
        return self._transposition

    @transposition.setter
    def transposition(self, val):
        self._transposition = val

    def get_fundamental_frequency(self):
        return midi_to_frequency(self.key, self.a4)

    def get_partial_midi_value(self, number):
        if not isinstance(number, int):
            return TypeError()
        if number <= 0:
            return ValueError()

        return frequency_to_midi(self.get_fundamental_frequency() * number, self.a4)


class StringInstrument(TreeInstrument):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.strings = {}


class Violin(StringInstrument):
    def __init__(self, number=None, *args, **kwargs):
        super().__init__(name='Violin', abbreviation='vln.', number=number, *args, **kwargs)
        self.strings = {4: String(4, G(3)),
                        3: String(3, D(4)),
                        2: String(2, A(4)),
                        1: String(1, E(5))
                        }
        self.standard_clef = TREBLE_CLEF


class Viola(StringInstrument):
    def __init__(self, number=None, *args, **kwargs):
        super().__init__(name='Viola', abbreviation='vla.', number=number, *args, **kwargs)
        self.strings = {4: String(4, C(3)),
                        3: String(3, G(3)),
                        2: String(2, D(4)),
                        1: String(1, A(4))
                        }
        self.standard_clef = ALTO_CLEF


class Cello(StringInstrument):
    def __init__(self, number=None, *args, **kwargs):
        super().__init__(name='Cello', abbreviation='vc.', number=number, *args, **kwargs)
        self.strings = {4: String(4, C(2)),
                        3: String(3, G(2)),
                        2: String(2, D(3)),
                        1: String(1, A(3))
                        }
        self.standard_clef = BASS_CLEF


class ViolaDamore(StringInstrument):
    def __init__(self, number=None, *args, **kwargs):
        super().__init__(name='Viola d\'amore\n430', abbreviation='vla.', number=number, *args, **kwargs)
        # skordatura
        self.strings = {1: String(1, B(4)),
                        2: String(2, B(4)),
                        3: String(3, F(4, '#')),
                        4: String(4, C(4)),
                        5: String(5, G(3)),
                        6: String(6, D(3)),
                        7: String(7, A(2))
                        }


class Accordion(TreeInstrument):
    def __init__(self, number=None, *args, **kwargs):
        super().__init__(name='Accordion', abbreviation='acc.', number=number, *args, **kwargs)


class Horn(TreeInstrument):
    def __init__(self, number=None, *args, **kwargs):
        super().__init__(name='Horn', abbreviation='hrn.', number=number, *args, **kwargs)


class NaturalHorn(NaturalInstrument):
    def __init__(self, key=E(1, 'b'), a4=430, *args, **kwargs):
        super().__init__(name='Horn in E♭\n430', abbreviation='hrn.', key=key, a4=a4, *args, **kwargs)
        self.transposition = 9


class TamTam(TreeInstrument):
    def __init__(self, number=None, *args, **kwargs):
        super().__init__(name='Tam-tam', abbreviation='Tam-t.', number=number, *args, **kwargs)
        self.midi = B(3)
        self.midi.notehead = 'x'


class Cymbal(TreeInstrument):
    midis = {1: E(4), 2: G(4), 3: B(4), 4: D(5), 5: F(5)}

    def __init__(self, number=1, *args, **kwargs):
        super().__init__(name='cymbal-' + str(number), abbreviation='cym-' + str(number), number=number, *args,
                         **kwargs)

        self.midi = self.midis[self.number]
        self.midi.notehead = 'x'


class Percussion(TreeInstrument):
    def __init__(self, number=None, *args, **kwargs):
        super().__init__(name='Percussion', abbreviation='perc.', number=number, *args, **kwargs)
        self.tamtam = TamTam()
        self.cymbal_1 = Cymbal(1)
        self.cymbal_2 = Cymbal(2)
        self.cymbal_3 = Cymbal(3)
        self.cymbal_4 = Cymbal(4)
        self.cymbal_5 = Cymbal(5)



from musicscore.musicxml.attributes.textformatting import TextFormatting
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import String


class ComplexTypeFormattedText(ComplexType, String, TextFormatting):
    """
    The formatted-text type represents a text element with text-formatting attributes.

    """

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Directive(AttributeAbstract):
    """
    The directive attribute changes the default-x position of a direction. It indicates that the left-hand side of
    the direction is aligned with the left-hand side of the time signature. If no time signature is present, it is
    aligned with the left-hand side of the first music notational element in the measure. If a default-x, justify,
    or halign attribute is present, it overrides the directive attribute.
    """

    def __init__(self, directive=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('directive', directive, 'TypeYesNo')



from unittest import TestCase
from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.elements.xml_element import XMLElement


class Positioned(XMLElement, Position):
    _ATTRIBUTES = ['default-x', 'default-y', 'relative-x', 'relative-y']

    def __init__(self, *args, **kwargs):
        super().__init__(tag='positioned', *args, **kwargs)


class PositionTest(TestCase):
    def setUp(self):
        self.positioned = Positioned()

    def test_position(self):
        result = '''<positioned/>
'''
        self.assertEqual(self.positioned.to_string(), result)

        self.positioned.default_x = 2
        result = '''<positioned default-x="2"/>
'''
        self.assertEqual(self.positioned.to_string(), result)

        self.positioned.default_y = 3
        result = '''<positioned default-x="2" default-y="3"/>
'''
        self.assertEqual(self.positioned.to_string(), result)

        self.positioned.default_x = None
        result = '''<positioned default-y="3"/>
'''
        self.assertEqual(self.positioned.to_string(), result)

        self.positioned.relative_y = 2
        result = '''<positioned default-y="3" relative-y="2"/>
'''
        self.assertEqual(self.positioned.to_string(), result)

        self.positioned.default_x = 5
        result = '''<positioned default-x="5" default-y="3" relative-y="2"/>
'''
        self.assertEqual(self.positioned.to_string(), result)



from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
import os

from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def test_1(self):
        score = TreeScoreTimewise()

        sf = SimpleFormat(midis=[(60, 64, 68, 73)], quarter_durations=[4])
        sf.to_stream_voice().add_to_score(score=score)

        sf.chords[0].inverse()
        sf.to_stream_voice().add_to_score(score=score, part_number=2)

        xml_path = path + '.xml'
        score.write(xml_path)
        TestScore().assert_template(xml_path)



from musicscore.dtd.dtd import Sequence, Element, Choice, GroupReference
from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import Empty
from musicscore.musicxml.types.simple_type import TypeStep, TypeSemitones, TypeOctave


class Step(XMLElement, TypeStep):
    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='step', value=value, *args, **kwargs)


class Alter(XMLElement, TypeSemitones):
    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='alter', value=value, *args, **kwargs)


class Octave(XMLElement, TypeOctave):
    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='octave', value=value, *args, **kwargs)


class DisplayStep(XMLElement, TypeStep):
    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='display-step', value=value, *args, **kwargs)


class DisplayOctave(XMLElement, TypeOctave):
    def __init__(self, value, *args, **kwargs):
        super().__init__(tag='display-octave', value=value, *args, **kwargs)


"""
The display-step-octave group contains the sequence of elements used by both the rest and unpitched elements. This
group is used to place rests and unpitched elements on the staff without implying that these elements have pitch.
Positioning follows the current clef. If percussion clef is used, the display-step and display-octave elements are
interpreted as if in treble clef, with a G in octave 4 on line 2. If not present, the note is placed on the middle
line"""
DisplayStepOctave = Sequence(
    Element(DisplayStep),
    Element(DisplayOctave)
)


class Event(XMLElement):
    """"""

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)


class Pitch(Event):
    """
    Pitch is represented as a combination of the step of the diatonic scale, the chromatic alteration, and the octave.
    """
    _DTD = Sequence(
        Element(Step),
        Element(Alter, min_occurrence=0),
        Element(Octave)
    )

    def __init__(self, step=Step('C'), alter=None, octave=Octave(4)):
        super().__init__(tag='pitch')
        self._step = None
        self.step = step
        self.alter = alter
        self._octave = None
        self.octave = octave

    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, value):
        self._set_child(Step, 'step', value)

    @property
    def alter(self):
        alter = self.get_children_by_type(Alter)
        if alter:
            return alter[0]
        else:
            return None

    @alter.setter
    def alter(self, value):
        self._set_child(Alter, 'alter', value)

    @property
    def octave(self):
        return self._octave

    @octave.setter
    def octave(self, value):
        self._set_child(Octave, 'octave', value)


class Unpitched(Event):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='unpitched', *args, **kwargs)


class Measure(AttributeAbstract):
    def __init__(self, measure=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('measure', measure, 'TypeYesNo')


class Rest(Event, Measure):
    """
    The rest element indicates notated rests or silences. Rest elements are usually empty, but placement on the staff
    can be specified using display-step and  display-octave elements. If the measure attribute is set to yes, this
    indicates this is a complete measure rest.
    """

    _DTD = Sequence(
        GroupReference(DisplayStepOctave, min_occurrence=0)
    )

    def __init__(self, measure=None, *args, **kwargs):
        super().__init__(tag='rest', measure=measure, *args, **kwargs)


class Chord(Empty):
    """
    The chord element indicates that this note is an additional chord tone with the preceding note. The duration of
    this note can be no longer than the preceding note. In MuseData, a missing duration indicates the same length as
    the previous note, but the MusicXML format requires a duration for chord notes too
    """

    def __init__(self, *args, **kwargs):
        super().__init__(tag='chord', *args, **kwargs)


"""
The full-note group is a sequence of the common note elements between cue/grace notes and regular (full) notes:
pitch, chord, and rest information, but not duration (cue and grace notes do not have duration encoded). Unpitched
elements are used for unpitched percussion, speaking voice, and other musical elements lacking determinate pitch.
"""

FullNote = Sequence(
    Element(Chord, min_occurrence=0),
    Choice(
        Element(Pitch),
        Element(Unpitched),
        Element(Rest)
    )
)



import copy

from musicscore.basic_functions import flatten


class Tree(object):
    """
    A simple Tree class
    """

    def __init__(self, label=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._children = []
        self._up = None
        self._leaves = []
        self.label = label

    @property
    def up(self):
        return self._up

    @property
    def is_root(self):
        if self._up is None:
            return True
        else:
            return False

    @property
    def index(self):
        if self.is_root:
            _index = [0]
        elif self.get_distance() == 1:
            _index = [self.up.get_children().index(self) + 1]
        else:
            _index = self.up.index
            _index.append(self.up.get_children().index(self) + 1)
        return _index

    def get_root(self):
        if self.is_root:
            return self

        root = self.up
        while not root.is_root:
            root = root.get_root()
        return root

    def get_children(self):
        return self._children

    def add_child(self, child):
        if not isinstance(child, Tree):
            raise TypeError('child must be of type Tree and not {}'.format(type(child)))
        self._children.append(child)
        child._up = self
        return child

    def remove_child(self, child):
        try:
            self._children.remove(child)
        except ValueError:
            pass

    def remove(self):
        if self.is_root:
            raise Exception()

        parent = self.up
        insert_index = self.up._children.index(self)
        self.up._children.remove(self)

        for child in self.get_children().__reversed__():
            parent.add_child(child)
            new_child = parent._children.pop(-1)
            parent._children.insert(insert_index, new_child)

    def clear_children(self):

        self._children.clear()

    def get_branch(self):
        output = [self]
        node = self
        while node.up is not None:
            output.append(node.up)
            node = node.up
        output.reverse()
        return output

    def get_common_ancestor(self, *other_nodes):
        for other_node in other_nodes:
            if self.get_root() != other_node.get_root():
                raise Exception('{} has not the same root'.format(other_node))

        me_branch = self.get_branch()[:]
        other_branches = [other_node.get_branch()[:] for other_node in other_nodes]

        for node in reversed(me_branch):
            node_in_all_branches = True
            for branch in other_branches:
                if node not in branch:
                    node_in_all_branches = False
                    break
            if node_in_all_branches:
                return node

    @property
    def is_leaf(self):
        if len(self.get_children()) == 0:
            return True
        else:
            return False

    def get_leaves(self, key=None):
        output = []
        for child in self.get_children():
            if not child.is_leaf:
                output.append(child.get_leaves(key=key))
            else:
                if key is not None:
                    output.append(key(child))
                else:
                    output.append(child)

        return output

    def traverse(self):
        yield self
        for child in self.get_children():
            for grand_child in child.traverse():
                yield grand_child

    def dump(self):
        output = []
        for node in self.traverse():
            output.append(node)
        return output

    def get_distance(self, reference=None):
        if reference is None:
            reference = self.get_root()

        if self.is_root:
            return 0
        parent = self.up
        count = 1
        while parent is not reference:
            parent = parent.up
            count += 1
            if parent.is_root and parent is not reference:
                return None
        return count

    def get_farthest_leaf(self):
        leaves = list(self.traverse_leaves())
        if not leaves:
            leaves = [self]
        return max(leaves, key=lambda leaf: leaf.get_distance())

    def get_layer(self, layer, key=None):
        if layer == 0:
            output = [self]
        elif layer == 1:
            output = self.get_children()
        else:
            output = []
            for child in self.get_layer(layer - 1):
                if child.is_leaf:
                    output.append(child)
                else:
                    output.extend(child.get_children())
        if key is None:
            return output
        else:
            return [key(child) for child in output]

    def get_siblings(self):
        if self.is_root:
            return []
        output = []
        for child in self.up.get_children():
            if child != self:
                output.append(child)
        return output

    def detach(self):
        self.up._children.remove(self)
        self._up = None

    def clone(self, except_nodes=[]):
        if self not in except_nodes:
            cloned = copy.copy(self)
            cloned._children = []
            for child in self.get_children():
                cloned.add_child(child.clone(except_nodes=except_nodes))
            return cloned
        else:
            return self

    def goto(self, index):

        if index == [0]:
            return self

        if len(index) == 1:
            return self.get_children()[index[0] - 1]

        return self.goto(index[:1]).goto(index[1:])

    def replace_node(self, new_node):
        if self.is_root:
            raise Exception('root cannot be replaced')
        else:
            index = self.up.get_children().index(self)
            new_node._up = self._up
            self.up.get_children()[index] = new_node
            self._up = None

    def traverse_leaves(self):
        for leaf in flatten(self.get_leaves()):
            yield leaf

    def find_leaf(self, condition):
        for leaf in self.traverse_leaves():
            if condition(leaf) is True:
                return leaf
        return False

    def __deepcopy__(self, *args, **kwargs):
        copied = self.__class__(*args, **kwargs)
        for item in self.__dict__.items():
            key = item[0]
            value = item[1]

            if key not in ('_children', '_up'):
                try:
                    new_value = value.__deepcopy__()
                except (AttributeError, TypeError):
                    new_value = value

                copied.__dict__[key] = new_value

        for child in self.get_children():
            copied.add_child(child.__deepcopy__())

        return copied

    def no_child_copy(self, *arguments, **kwargs):
        # Tree.COUNT_COPYING += 1
        # print(Tree.COUNT_COPYING)
        # print(self)
        copied = self.__class__(*arguments, **kwargs)
        for item in self.__dict__.items():
            key = item[0]
            value = item[1]

            if key not in ('_children', '_up'):
                try:
                    new_value = value.__copy__()
                except (AttributeError, TypeError):
                    new_value = value

                copied.__dict__[key] = new_value

        return copied

    @property
    def __next__(self):
        if not self.is_leaf:
            raise Exception('__next__ is only possible for leaves')

        leave_iterator = self.get_root().traverse_leaves()

        def forward_leaves(leave_iterator):
            leaf = leave_iterator.__next__()
            if leaf != self:
                forward_leaves(leave_iterator)

        forward_leaves(leave_iterator)
        return leave_iterator.__next__()


class TreeNode(Tree):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract

class BezierX(AttributeAbstract):
    def __init__(self, bezier_x=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('bezier-x', bezier_x, "TypeTenths")


class BezierY(AttributeAbstract):
    def __init__(self, bezier_y=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('bezier-y', bezier_y, "TypeTenths")


class BezierX2(AttributeAbstract):
    def __init__(self, bezier_x2=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('bezier-x2', bezier_x2, "TypeTenths")


class BezierY2(AttributeAbstract):
    def __init__(self, bezier_y2=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('bezier-y2', bezier_y2, "TypeTenths")


class BezierOffset(AttributeAbstract):
    def __init__(self, bezier_offset=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('bezier-offset', bezier_offset, "TypeDivisions")


class BezierOffset2(AttributeAbstract):
    def __init__(self, bezier_offset2=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('bezier-offset2', bezier_offset2, "TypeDivisions")


class Bezier(BezierX, BezierY, BezierX2, BezierY2, BezierOffset, BezierOffset2):
    """
    The bezier attribute group is used to indicate the curvature of slurs and ties, representing the control points for
    a cubic bezier curve. For ties, the bezier attribute group is used with the tied element.

    Normal slurs, S-shaped slurs, and ties need only two bezier points: one associated with the start of the slur or
    tie, the other with the stop. Complex slurs and slurs divided over system breaks can specify additional bezier data
    at slur elements with a continue type.

    The bezier-x, bezier-y, and bezier-offset attributes describe the outgoing bezier point for slurs and ties with a
    start type, and the incoming bezier point for slurs and ties with types of stop or continue. The bezier-x2,
    bezier-y2, and bezier-offset2 attributes are only valid with slurs of type continue, and describe the outgoing
    bezier point.

    The bezier-x, bezier-y, bezier-x2, and bezier-y2 attributes are specified in tenths, relative to any position
    settings associated with the slur or tied element. The bezier-offset and bezier-offset2 attributes are measured in
    terms of musical divisions, like the offset element.

    The bezier-offset and bezier-offset2 attributes are deprecated as of MusicXML 3.1. If both the bezier-x and
    bezier-offset attributes are present, the bezier-x attribute takes priority. Similarly, the bezier-x2 attribute
    takes priority over the bezier-offset2 attribute. The two types of bezier attributes are not additive.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from unittest import TestCase
from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.elements.xml_element import XMLElement


class AttributeSub(XMLElement, AttributeAbstract):
    def __init__(self):
        super().__init__(tag='attribute-sub')
        self._ATTRIBUTES.append('print-object')


class TestAttributeAbstract(TestCase):
    def setUp(self):
        self.attribute_sub = AttributeSub()

    def test_attribute_abstract(self):
        self.attribute_sub.generate_attribute(attribute_name='print-object', attribute_value=2, type_string='TypeTenths')
        self.assertEqual(self.attribute_sub.print_object, 2)
        with self.assertRaises(TypeError):
            self.attribute_sub.print_object = 'b'



import warnings

from musicscore.tree.tree import Tree


class DTDError(Exception):
    def __init__(self, msg='DTD Error'):
        super().__init__(msg)


class ChildTypeDTDConflict(DTDError):
    def __init__(self, child, parent):
        msg = 'child of type {} cannot be added to {} due to DTD Type conflicts'.format(type(child), parent.__class__)
        super().__init__(msg)


class ChildOccurrenceDTDConflict(DTDError):
    def __init__(self, child, parent):
        msg = 'child of type {} cannot be added to {} due to DTD Occurrence conflicts'.format(type(child),
                                                                                              parent.__class__)
        super().__init__(msg)


class ChildIsNotOptional(DTDError):
    def __init__(self, node, parent):
        msg = 'child of type {} is due to DTD Occurrence not optional for {}'.format(node.type_.__name__,
                                                                                     parent.__class__)
        super().__init__(msg)


class DTDConflict(DTDError):
    def __init__(self, msg=None):
        if msg is None:
            msg = 'DTD conflicts exist'
        super().__init__(msg)


class DTDTree(Tree):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class DTDNode(DTDTree):
    def __init__(self, children=None, min_occurrence=1, max_occurrence=1, *args, **kwargs):
        super().__init__(**kwargs)
        self.choices = None
        self._min_occurrence = None
        self._max_occurrence = None
        self._choice_index = 0
        self._current_choice = None
        self._xml_children = None
        self.min_occurrence = min_occurrence
        self.max_occurrence = max_occurrence

        if children is None:
            children = []
        for child in children:
            if not isinstance(child, DTDTree):
                raise TypeError(
                    '{} can only have children of Type DTDTree not {}'.format(self.__class__.__name__, type(child)))
            self.add_child(child)

    @property
    def min_occurrence(self):
        return self._min_occurrence

    @min_occurrence.setter
    def min_occurrence(self, value):
        if value is not None and not isinstance(value, int):
            raise TypeError('min_occurrence.value must be of type None or positive int not {}'.format(type(value)))
        if value is not None and value < 0:
            raise ValueError('min_occurrence.value must be positive'.format(type(value)))
        self._min_occurrence = value

    @property
    def max_occurrence(self):
        return self._max_occurrence

    @max_occurrence.setter
    def max_occurrence(self, value):
        if value is not None and not isinstance(value, int):
            raise TypeError('max_occurrence.value must be of type None or positive int not {}'.format(type(value)))
        if value is not None and value < 0:
            raise ValueError('max_occurrence.value must be positive'.format(type(value)))
        self._max_occurrence = value

    @property
    def xml_children(self):
        if not self._xml_children:
            self._xml_children = []
        return self._xml_children

    def __repr__(self):
        if isinstance(self, Element):
            return '{} type {} at {}'.format(self.__class__.__name__, self.type_.__name__, hex(id(self)))
        return '{} at {}'.format(self.__class__.__name__, hex(id(self)))

    def copy_child(self, child, ch=None):
        if isinstance(child, Element):
            copied = child.no_child_copy(child.type_)
        else:
            copied = child.no_child_copy()

        copied.original = child
        copied.choice = ch
        return copied

    def eliminate_group_reference(self):

        for node in self.traverse():
            if isinstance(node, GroupReference):
                if len(node.get_children()) > 1:
                    warnings.warn(
                        'dtd.eliminate_group_reference: group has more than one child! Only the first child will be '
                        'replaced.')
                new_node = node.get_children()[0]
                if (new_node.min_occurrence, new_node.max_occurrence) != (1, 1):
                    warnings.warn('dtd.eliminate_group_reference: groups child occurrences are not 1.')
                (new_node.min_occurrence, new_node.max_occurrence) = (node.min_occurrence, node.max_occurrence)
                node.replace_node(new_node)

    def carbon_copy(self):
        if isinstance(self, Element):
            copied = Element(self.type_)
        else:
            copied = self.__class__()

        copied.original = self.original
        copied.choice = self.choice
        (copied.min_occurrence, copied.max_occurrence) = (self.min_occurrence, self.max_occurrence)
        for child in self.get_children():
            copied.add_child(child.carbon_copy())

        return copied

    def generate_dtd_choices(self):
        self.eliminate_group_reference()

        self.choices = [self.copy_child(self)]
        choice_index = 0

        while choice_index < len(self.choices):
            current_choice = self.choices[choice_index]
            current_choice.magic_expand(self.choices)
            choice_index += 1

    def get_choices(self):
        if not self.choices:
            self.generate_dtd_choices()
        return self.choices

    def remove_xml_child(self, xml_child):
        for node in self.traverse():
            if xml_child in node.xml_children:
                node.xml_children.remove(xml_child)

    def get_xml_children(self):
        xml_children = []
        pregnant_nodes = []

        for leaf in self.traverse_leaves():
            if isinstance(leaf.up, Choice) and (leaf.up.min_occurrence, leaf.up.max_occurrence) == (0, None):
                if leaf.up not in pregnant_nodes:
                    pregnant_nodes.append(leaf.up)
            else:
                pregnant_nodes.append(leaf)
        for node in pregnant_nodes:
            xml_children.extend(node.xml_children)
        return xml_children


class DTDLeaf(DTDNode):
    def __init__(self, type_, min_occurrence=1, max_occurrence=1, *args, **kwargs):
        super().__init__(min_occurrence=min_occurrence, max_occurrence=max_occurrence, *args, **kwargs)
        self._type = None
        self.type_ = type_

    @property
    def type_(self):
        return self._type

    @type_.setter
    def type_(self, value):
        if not isinstance(value, type):
            raise TypeError('type_.value must be of type type not{}'.format(type(value)))
        self._type = value

    def add_child(self, child):
        raise Exception('DTDLeaf can have no children')


class Choice(DTDNode):
    """"""

    def __init__(self, *children, min_occurrence=1, max_occurrence=1):
        super().__init__(children=children, min_occurrence=min_occurrence, max_occurrence=max_occurrence)

    def magic_expand(self, dtd_choices):
        if not self.original:
            raise Exception('is not copied')

        if (self.min_occurrence, self.max_occurrence) == (1, 1):
            if self.choice is None:
                for index, child in enumerate(self.original.get_children()):
                    if index == 0:
                        copied = self.copy_child(child, index)
                        self.add_child(copied)
                        self.choice = 0
                        copied.magic_expand(dtd_choices)
                    else:
                        new_choice = self.get_root().carbon_copy()
                        new_choice.goto(self.index).choice = index
                        new_choice.goto(self.index)._children = []
                        dtd_choices.append(new_choice)
            else:
                if not self.get_children():
                    copied = self.copy_child(self.original.get_children()[self.choice])
                    self.add_child(copied)
                    copied.magic_expand(dtd_choices)
                else:
                    self.get_children()[0].magic_expand(dtd_choices)
        else:
            for child in self.original.get_children():
                child_exists = False
                for ch in self.get_children():
                    if ch.original == child:
                        child_exists = True
                        ch.magic_expand(dtd_choices)
                        break

                if not child_exists:
                    copied = self.copy_child(child)
                    self.add_child(copied)
                    copied.magic_expand(dtd_choices)


class Sequence(DTDNode):
    """"""

    def __init__(self, *children, min_occurrence=1, max_occurrence=1):
        super().__init__(children=children, min_occurrence=min_occurrence, max_occurrence=max_occurrence)

    def magic_expand(self, dtd_choices):
        if not self.original:
            raise Exception('is not a copy')
        for child in self.original.get_children():
            child_exists = False
            for ch in self.get_children():
                if ch.original == child:
                    child_exists = True
                    ch.magic_expand(dtd_choices)
                    break

            if not child_exists:
                copied = self.copy_child(child)
                self.add_child(copied)
                copied.magic_expand(dtd_choices)


class Element(DTDLeaf):
    """"""

    def __init__(self, type_, min_occurrence=1, max_occurrence=1, **kwargs):
        super().__init__(type_, min_occurrence=min_occurrence, max_occurrence=max_occurrence, **kwargs)

    def magic_expand(self, dtd_choices):
        pass

    def __deepcopy__(self):
        return Element(type_=self.type_, min_occurrence=self.min_occurrence, max_occurrence=self.max_occurrence)

    def check_min_occurrence(self):
        if len(self.xml_children) >= self.min_occurrence:
            return True

        if isinstance(self.up, Choice) and self.up.min_occurrence == 0:
            return True

        if isinstance(self.up, Sequence) and self.up.min_occurrence == 0:
            siblings = [child for child in self.up.get_children() if child != self]
            for sibling in siblings:
                if sibling._xml_children:
                    return False
            return True

        if self.up.min_occurrence != 1:
            raise NotImplementedError('check_min_occurrence')
        else:
            return False

    def add_xml_child(self, xml_child):
        if not isinstance(xml_child, self.type_):
            raise DTDError('{} is of wrong type and cannot be added to {}'.format(xml_child, self))

        if isinstance(self.up, Choice) and (self.up.min_occurrence, self.up.max_occurrence) == (0, None):
            self.up.xml_children.append(xml_child)
            xml_child.node = self.up
            return True

        elif self.max_occurrence is None or len(self.xml_children) < self.max_occurrence:
            self.xml_children.append(xml_child)
            xml_child.node = self
            return True

        elif self.up.max_occurrence != 1:
            if isinstance(self.up, Sequence):
                return False
            else:
                # self.up.xml_children.append(xml_child)
                # xml_child.node = self.up
                self.xml_children.append(xml_child)
                xml_child.node = self
                return True
                # raise NotImplementedError()

        else:
            return False


class GroupReference(DTDNode):
    """"""

    def __init__(self, child, min_occurrence=1, max_occurrence=1):
        super().__init__(children=[child.__deepcopy__()], min_occurrence=min_occurrence, max_occurrence=max_occurrence)

    @property
    def child(self):
        return self.get_children()[0]

    def __deepcopy__(self):
        return GroupReference(self.child.__deepcopy__(), min_occurrence=self.min_occurrence,
                              max_occurrence=self.max_occurrence)



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.types.complextypes.complextype import ComplexType, EmptyPrintObjectStyleAlign


class LeftDivider(EmptyPrintObjectStyleAlign):
    _TAG = 'left-divider'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class RightDivider(EmptyPrintObjectStyleAlign):
    _TAG = 'right-divider'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypeSystemDividers(ComplexType):
    """The system-dividers element indicates the presence or absence of system dividers (also known as system
    separation marks) between systems displayed on the same page. Dividers on the left and right side of the page are
    controlled by the left-divider and right-divider elements respectively. The default vertical position is half the
    system-distance value from the top of the system that is below the divider. The default horizontal position is the
    left and right system margin, respectively.

    When used in the print element, the system-dividers element affects the dividers that would appear between the
    current system and the previous system."""

    _DTD = Sequence(
        Element(LeftDivider),
        Element(RightDivider),
    ),

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeHoleClosedValue, TypeHoleClosedLocation


class ComplexTypeHoleClosed(ComplexType, TypeHoleClosedValue):
    """The hole-closed type represents whether the hole is closed, open, or half-open. The optional location attribute
    indicates which portion of the hole is filled in when the element value is half.
    """

    def __init__(self, tag, value, location=None, *args, **kwargs):
        super().__init__(tag=tag, value=value, *args, **kwargs)
        self.location = location

    @property
    def location(self):
        return self.get_attribute('location')

    @location.setter
    def location(self, value):
        if value is None:
            self.remove_attribute('location')
        else:
            TypeHoleClosedLocation(value)
            self._ATTRIBUTES.insert(0, 'location')
            self.set_attribute('location', value)



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeTenths, TypeStaffNumber


class StaffDistance(XMLElement, TypeTenths):
    _TAG = 'staff-distance'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class ComplexTypeStaffLayout(ComplexType):
    """
    Staff layout includes the vertical distance from the bottom line of the previous staff in this system to the top
    line of the staff specified by the number attribute. The optional number attribute refers to staff numbers within
    the part, from top to bottom on the system. A value of 1 is assumed if not present. When used in the defaults
    element, the values apply to all parts. This value is ignored for the first staff in a system.
    """
    _DTD = Sequence(
        Element(StaffDistance,
                min_occurrence=0)

    )

    def __init__(self, tag, number=None, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.number = number

    @property
    def number(self):
        return self.get_attribute('number')

    @number.setter
    def number(self, value):
        if value is None:
            self.remove_attribute('number')
        else:
            TypeStaffNumber(value)
            self._ATTRIBUTES.insert(0, 'number')
            self.set_attribute('number', value)



from musicscore.dtd.dtd import Sequence, Element, Choice
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.partgroup import ComplexTypePartGroup
from musicscore.musicxml.types.complextypes.scorepart import ComplexTypeScorePart


class ScorePart(ComplexTypeScorePart):
    """
    The score-part element is defined within a group due to its multiple uses within the part-list element.
    Each MusicXML part corresponds to a track in a Standard MIDI Format 1 file. The score-instrument elements are used
    when there are multiple instruments per track. The midi-device element is used to make a MIDI device or port
    assignment for the given track. Initial midi-instrument assignments may be made here as well.
    """
    _TAG = 'score-part'

    def __init__(self, id, *args, **kwargs):
        super().__init__(tag=self._TAG, id=id, *args, **kwargs)


class PartGroup(ComplexTypePartGroup):
    _TAG = 'part-group'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class ComplexTypePartList(ComplexType):
    """
    The part-list identifies the different musical parts in this movement. Each part has an ID that is used later within
    the musical data. Since parts may be encoded separately and combined later, identification elements are present at
    both the score and score-part levels. There must be at least one score-part, combined as desired with part-group
    elements that indicate braces and brackets. Parts are ordered from top to bottom in a score based on the order in
    which they appear in the part-list.
    """
    # _DTD = Sequence(
    #     Element(PartGroup, min_occurrence=0, max_occurrence=None),
    #     Element(ScorePart),
    #     Choice(
    #         Element(PartGroup),
    #         Element(ScorePart),
    #         min_occurrence=0, max_occurrence=None
    #     )
    # )
    _DTD = Sequence(
        Choice(Element(PartGroup, min_occurrence=0, max_occurrence=None),
               Element(ScorePart)
               ),
        Choice(
            Element(PartGroup),
            Element(ScorePart),
            min_occurrence=0, max_occurrence=None
        )
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Element(AttributeAbstract):
    def __init__(self, element=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('element', element, "Token")


class ElementPosition(Element, Position):
    """
    The element and position attributes are new as of Version 2.0. They allow for bookmarks and links to be positioned
    at higher resolution than the level of music-data elements. When no element and position attributes are present,
    the bookmark or link element refers to the next sibling element in the MusicXML file. The element attribute
    specifies an element type for a descendant of the next sibling element that is not a link or bookmark. The position
    attribute specifies the position of this descendant element, where the first position is 1. The position attribute
    is ignored if the element attribute is not present. For instance, an element value of "beam" and a position value
    of "2" defines the link or bookmark to refer to the second beam descendant of the next sibling element that is not
    a link or bookmark. This is equivalent to an XPath test of [.//beam[2]] done in the context of the sibling element.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from musicscore.dtd.dtd import GroupReference, Element, Sequence

from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.simple_type import TypeTenths


class LeftMargin(XMLElement, TypeTenths):
    _TAG = 'left-margin'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class RightMargin(XMLElement, TypeTenths):
    _TAG = 'right-margin'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class TopMargin(XMLElement, TypeTenths):
    _TAG = 'top-margin'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class BottomMargin(XMLElement, TypeTenths):
    _TAG = 'bottom-margin'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


"""
The left-right-margins group specifies horizontal margins in tenths.
"""

LeftRightMargins = Sequence(
    Element(LeftMargin),
    Element(RightMargin)
)

"""
The all-margins group specifies both horizontal and vertical margins in tenths.
"""
AllMargins = Sequence(
    GroupReference(LeftRightMargins),
    Element(TopMargin),
    Element(BottomMargin)
)



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeMillimeters, TypeTenths


class Millimeters(XMLElement, TypeMillimeters):
    _TAG = 'millimeters'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Tenths(XMLElement, TypeTenths):
    _TAG = 'tenths'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class ComplexTypeScaling(ComplexType):
    """
    Margins, page sizes, and distances are all measured in tenths to keep MusicXML data in a consistent coordinate
    system as much as possible. The translation to absolute units is done with the scaling type, which specifies how
    many millimeters are equal to how many tenths. For a staff height of 7mm, millimeters would be set to 7 while tenths
    is set to 40. The ability to set a formula rather than a single scaling factor helps avoid roundoff errors.
    """
    _DTD = Sequence(
        Element(Millimeters),
        Element(Tenths)
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)


from musicscore.musicxml.attributes.font import Font
from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.elements.fullnote import Pitch


from unittest import TestCase


class XMLWithFont(Pitch, Font, Position):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TestFont(TestCase):
    def setUp(self):
        self.font = XMLWithFont()

    def test_font(self):

        self.font.font_weight = 'bold'
        self.font.font_size = 8

        with self.assertRaises(TypeError):
            self.font.font_size = 'b'

        self.font.font_style = 'italic'
        self.font.font_family = 'Arial, Times'

        result = '''<pitch font-family="Arial, Times" font-style="italic" font-size="8" font-weight="bold">
  <step>C</step>
  <octave>4</octave>
</pitch>
'''
        self.assertEqual(self.font.to_string(), result)



from musicscore.musicxml.types.simple_type import TypeCommaSeparatedText
from unittest import TestCase


class TestComma(TestCase):
    def test_comma(self):
        comma = TypeCommaSeparatedText('bla,  bla, bla')
        with self.assertRaises(ValueError):
            comma = TypeCommaSeparatedText(', bla, bla')
        with self.assertRaises(ValueError):
            comma = TypeCommaSeparatedText('bla, bla,')




import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.elements.note import Notations, Duration
from musicscore.musicxml.types.complextypes.dynamics import FF, Dynamics
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):

    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        chord = TreeChord()
        chord.add_dynamics('ff')
        self.assertEqual(chord.get_dynamics(), 'ff')





import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = str(os.path.abspath(__file__).split('.')[0])

score = TreeScoreTimewise()
simple_format = SimpleFormat(quarter_durations=[1, 4])
simple_format.chords[-1].add_grace_chords(TreeChord(midis=[66]), mode='post')
simple_format.to_stream_voice().add_to_score(score)

xml_path = path + '.xml'
score.write(xml_path)



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        midis = [(60, 63, 65), 80]
        sf = SimpleFormat(midis=midis)
        voice = sf.to_stream_voice(1)
        l = voice.chords[0].add_lyric('bla')

        voice.add_to_score(self.score, 1, 1)
        result_path = path + '_test_1'
        self.score.write(result_path)
        TestScore().assert_template(result_path=result_path)

    def test_2(self):
        midis = [(60, 63, 65), 80]
        sf = SimpleFormat(midis=midis)
        voice = sf.to_stream_voice(1)
        voice.chords[0].add_lyric('bla')
        voice.chords[0].add_lyric('bb', number=2)

        voice.add_to_score(self.score, 1, 1)
        result_path = path + '_test_2'
        self.score.write(result_path)
        TestScore().assert_template(result_path=result_path)

    def test_3(self):
        midis = [(60, 63, 65), 80]
        sf = SimpleFormat(midis=midis)
        voice1 = sf.to_stream_voice(2)
        voice1.chords[0].add_lyric('bla')
        voice1.chords[0].add_lyric('bb', number=2)
        voice1.add_to_score(self.score, 1, 1)

        midis = [90, (91, 92)]
        sf = SimpleFormat(midis=midis)
        voice2 = sf.to_stream_voice(1)
        voice2.chords[1].add_lyric('sh')
        voice2.chords[1].add_lyric('th', number=2)
        voice2.add_to_score(self.score, 1, 1)

        result_path = path + '_test_3'
        self.score.write(result_path)
        TestScore().assert_template(result_path=result_path)



from unittest import TestCase

from musicscore.musicxml.elements.note import Type


class TestNoteType(TestCase):
    def setUp(self):
        self.type = Type('quarter')

    def test_type(self):
        result = '''<type>quarter</type>
'''
        self.assertEqual(self.type.to_string(), result)

        with self.assertRaises(ValueError):
            Type('bla')



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract, TypeYesNo, TypeTrillBeats, TypePercent

class Accelerate(AttributeAbstract):
    def __init__(self, accelerate=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('accelerate', accelerate, "TypeYesNo")
        TypeYesNo


class Beats(AttributeAbstract):
    def __init__(self, beats=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('beats', beats, "TypeTrillBeats")
        TypeTrillBeats


class FirstBeat(AttributeAbstract):
    def __init__(self, first_beat=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('first-beat', first_beat, "TypePercent")
        TypePercent


class LastBeat(AttributeAbstract):
    def __init__(self, last_beat=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('last-beat', last_beat, "TypePercent")
        TypePercent


class BendSound(Accelerate, Beats, FirstBeat, LastBeat):
    """
    The bend-sound type is used for bend and slide elements, and is similar to the trill-sound attribute group. Here
    the beats element refers to the number of discrete elements (like MIDI pitch bends) used to represent a continuous
    bend or slide. The first-beat indicates the percentage of the direction for starting a bend; the last-beat the
    percentage for ending it. The default choices are
    accelerate = "no"
    beats = "4"
    first-beat = "25"
    last-beat = "75"
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from unittest import TestCase

from musicscore.musicxml.groups.common import FootNote
from musicscore.musicxml.types.complextypes.lyric import ComplexTypeLyric, Extend, Syllabic, Text, EndLine


class Test(TestCase):
    def setUp(self):
        self.lyric = ComplexTypeLyric()
        self.lyric.add_child(Extend())
        self.lyric.add_child(Syllabic('begin'))
        self.lyric.add_child(Text('bla'))
        # self.lyric.add_child(Elision('00A0'))
        # self.lyric.add_child(Elision('00A0'))
        self.lyric.add_child(EndLine())
        self.lyric.add_child(FootNote('foot'))

    def test_1(self):
        result = self.lyric.to_string()
        expected = '''<lyric number="1">
  <syllabic>begin</syllabic>
  <text>bla</text>
  <extend/>
  <end-line/>
  <footnote>foot</footnote>
</lyric>
'''
        self.assertEqual(expected, result)


from unittest import TestCase

from musicscore.musicxml.elements.scoreheader import PartList
from musicscore.musicxml.types.complextypes.partlist import ScorePart, PartGroup
from musicscore.musicxml.types.complextypes.scorepart import PartName


def generate_score_part(id):
    output = ScorePart(id=id)
    output.add_child(PartName(id))
    return output


class Test(TestCase):

    def test_1(self):
        pl = PartList()
        pl.add_child(generate_score_part('p1'))
        pl.add_child(generate_score_part('p2'))
        result = pl.to_string()
        expected = '''<part-list>
  <score-part id="p1">
    <part-name>p1</part-name>
  </score-part>
  <score-part id="p2">
    <part-name>p2</part-name>
  </score-part>
</part-list>
'''
        self.assertEqual(expected, result)

    def test_2(self):
        pl = PartList()
        pl.add_child(PartGroup(type='start'))
        pl.add_child(generate_score_part('p1'))
        pl.add_child(generate_score_part('p2'))
        result = pl.to_string()
        expected = '''<part-list>
  <part-group number="1" type="start"/>
  <score-part id="p1">
    <part-name>p1</part-name>
  </score-part>
  <score-part id="p2">
    <part-name>p2</part-name>
  </score-part>
</part-list>
'''
        self.assertEqual(expected, result)

    def test_3(self):
        pl = PartList()
        pl.add_child(PartGroup(type='start'))
        pl.add_child(generate_score_part('p1'))
        pl.goto_next_dtd_choice()
        pl.add_child(generate_score_part('p2'))
        pl.add_child(PartGroup(type='stop'))
        result = pl.to_string()

        expected = '''<part-list>
  <score-part id="p1">
    <part-name>p1</part-name>
  </score-part>
  <part-group number="1" type="start"/>
  <score-part id="p2">
    <part-name>p2</part-name>
  </score-part>
  <part-group number="1" type="stop"/>
</part-list>
'''
        self.assertEqual(expected, result)

    def test_4(self):
        pl = PartList()
        pl.add_child(generate_score_part('p1'))
        pl.goto_next_dtd_choice()
        pl.add_child(PartGroup(type='start'))
        pl.add_child(generate_score_part('p2'))
        pl.add_child(generate_score_part('p3'))
        pl.add_child(generate_score_part('p4'))
        pl.add_child(PartGroup(type='stop'))
        pl.add_child(generate_score_part('p5'))
        result = pl.to_string()
        expected = '''<part-list>
  <score-part id="p1">
    <part-name>p1</part-name>
  </score-part>
  <part-group number="1" type="start"/>
  <score-part id="p2">
    <part-name>p2</part-name>
  </score-part>
  <score-part id="p3">
    <part-name>p3</part-name>
  </score-part>
  <score-part id="p4">
    <part-name>p4</part-name>
  </score-part>
  <part-group number="1" type="stop"/>
  <score-part id="p5">
    <part-name>p5</part-name>
  </score-part>
</part-list>
'''
        self.assertEqual(expected, result)



from unittest import TestCase

from musicscore.musicxml.elements.fullnote import Rest
from musicscore.musicxml.groups.musicdata import Direction, Attributes
from musicscore.musicxml.elements.note import Note, Duration
from musicscore.musicxml.elements.timewise import Part
from musicscore.musicxml.types.complextypes.attributes import Divisions
from musicscore.musicxml.types.complextypes.direction import DirectionType
from musicscore.musicxml.types.complextypes.directiontype import Words

# todo: Symbol in DTD

class Test(TestCase):
    def test_1(self):
        part = Part(id='P1')
        attributes = part.add_child(Attributes())
        attributes.add_child(Divisions(2))
        d = part.add_child(Direction())
        dt = d.add_child(DirectionType())
        dt.add_child(Words(value='bla'))
        note = part.add_child(Note())
        note.add_child(Rest())
        note.add_child(Duration(8))

        result = '''<part id="P1">
  <attributes>
    <divisions>2</divisions>
  </attributes>
  <direction>
    <direction-type>
      <words>bla</words>
    </direction-type>
  </direction>
  <note>
    <rest/>
    <duration>8</duration>
  </note>
</part>
'''
        self.assertEqual(part.to_string(), result)



from unittest import TestCase

from musicscore.musicxml.groups.musicdata import Direction
from musicscore.musicxml.types.complextypes.direction import DirectionType
from musicscore.musicxml.types.complextypes.directiontype import Dynamics
from musicscore.musicxml.types.complextypes.dynamics import P


class Test(TestCase):
    def test_1(self):
        direction = Direction()
        direction_type = direction.add_child(DirectionType())
        dynamics = direction_type.add_child(Dynamics())
        dynamics.add_child(P())
        result = """<direction>
  <direction-type>
    <dynamics placement="below">
      <p/>
    </dynamics>
  </direction-type>
</direction>
"""
        self.assertEqual(direction.to_string(), result)




from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Cautionary(AttributeAbstract):
    def __init__(self, cautionary=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('cautionary', cautionary, 'TypeYesNo')


class Editorial(AttributeAbstract):
    def __init__(self, editorial=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('editorial', editorial, 'TypeYesNo')


class Smulf(AttributeAbstract):
    def __init__(self, smulf=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('smulf', smulf, 'SmulfAccidentalGlyphName')



from musicscore.musicxml.attributes.elementposition import ElementPosition
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import ID, Token


class ComplexTypeBookmark(ComplexType, ElementPosition):
    """The bookmark type serves as a well-defined target for an incoming simple XLink."""

    def __init__(self, id_, name=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = id_
        self.name = name

    @property
    def id(self):
        return self.get_attribute('id')

    @id.setter
    def id(self, value):
        if value is None:
            self.remove_attribute('id')
        else:
            ID(value)
            self._ATTRIBUTES.insert(0, 'id')
            self.set_attribute('id', value)

    @property
    def name(self):
        return self.get_attribute('name')

    @name.setter
    def name(self, value):
        if value is None:
            self.remove_attribute('name')
        else:
            Token(value)
            self._ATTRIBUTES.insert(0, 'name')
            self.set_attribute('name', value)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class FontFamily(AttributeAbstract):
    """"""

    def __init__(self, font_family=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('font-family', font_family, 'TypeCommaSeparatedText')


class FontStyle(AttributeAbstract):
    """"""

    def __init__(self, font_style=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('font-style', font_style, 'TypeFontStyle')


class FontSize(AttributeAbstract):
    """"""

    def __init__(self, font_size=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('font-size', font_size, 'TypeFontSize')


class FontWeight(AttributeAbstract):
    """"""

    def __init__(self, font_weight=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.generate_attribute('font-weight', font_weight, 'TypeFontWeight')


class Font(FontFamily, FontStyle, FontSize, FontWeight):
    """	The font entity gathers together attributes for determining the font within a directive or direction.
    They are based on the text styles for Cascading Style Sheets. The font-family is a comma-separated list
    of font names. These can be specific font styles such as Maestro or Opus, or one of several generic font styles:
    music, engraved, handwritten, text, serif, sans-serif, handwritten, cursive, fantasy, and monospace. The music,
    engraved, and handwritten values refer to music fonts; the rest refer to text fonts. The fantasy style refers to
    decorative text such as found in older German-style printing. The font-style can be normal or italic. The
    font-size can be one of the CSS sizes (xx-small, x-small, small, medium, large, x-large, xx-large) or a numeric
    point size. The font-weight can be normal or bold. The default is application-dependent, but is a text font vs. """

    def __init__(self, font_weight=None, font_size=None, font_style=None, font_family=None, *args, **kwargs):
        super().__init__(font_weight=font_weight, font_size=font_size, font_style=font_style,
                         font_family=font_family, *args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Halign(AttributeAbstract):
    """
    In cases where text extends over more than one line, horizontal alignment and justify values can be different. The
    most typical case is for credits, such as:

	Words and music by
	  Pat Songwriter

	Typically this type of credit is aligned to the right, so that the position information refers to the right-most
	part of the text. But in this example, the text is center-justified, not right-justified.

	The halign attribute is used in these situations. If it is not present, its value is the same as for the justify
	attribute.
    """

    def __init__(self, halign=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('halign', halign, 'TypeLeftCenterRight')


class Valign(AttributeAbstract):
    """
    The valign attribute is used to indicate vertical alignment to the top, middle, bottom, or baseline of the text.
    Defaults are implementation-dependent.
    """

    def __init__(self, valign=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('valign', valign, 'TypeValign')


class ValignImage(AttributeAbstract):
    """
    The valign-image attribute is used to indicate vertical alignment for images and graphics, so it removes the
    baseline value. Defaults are implementation-dependent.
    """

    def __init__(self, valign_image=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('valign-image', valign_image, 'TypeValignImage')



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.types.complextypes.complextype import ComplexType


class Type(AttributeAbstract):
    def __init__(self, type=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('type', type, 'TypeStartStopContinue')


class ComplexTypeExtend(ComplexType, Type, Position, Color):
    """
    The extend type represents lyric word extension / melisma lines as well as figured bass extensions. The optional
    type and position attributes are added in Version 3.0 to provide better formatting control.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treebeat import TreeBeat
from musicscore.musictree.treechordflags import PizzFlag, PercussionFlag
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treepart import TreePart
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.elements.note import Type
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        # sf = SimpleFormat(durations=[4, 4, 2, 1, 1.5, 1.8, 0.2, 0.4, 0.5, 1])
        sf = SimpleFormat(quarter_durations=[4, 4])

        for chord in sf.chords:
            # chord.add_flag(PizzFlag())
            chord.add_flag(PercussionFlag())
        v = sf.to_stream_voice()
        self.score.set_time_signatures([4, 3, 1])
        v.add_to_score(self.score, 1, 1)
        result_path = path + '_test_1'

        # self.score.fill_with_rest()
        # self.score.add_beats()
        # self.score.quantize()
        # for measure in self.score.get_children_by_type(TreeMeasure):
        #     for part in measure.get_children_by_type(TreePart):
        #         for beat in part.get_beats():
        #             new_chords = []
        #             for chord in beat.chords:
        #                 if chord.is_tied_to_previous:
        #                     chord.to_rest()
        #                     new_chords.append(chord)
        #
        #                 elif chord.position_in_beat == 0:
        #                     split = [chord]
        #                     if chord.quarter_duration == 1:
        #                         split = chord.split(1, 1)
        #                     elif chord.quarter_duration == 2:
        #                         split = chord.split(1, 3)
        #                     elif chord.quarter_duration == 3:
        #                         split = chord.split(1, 5)
        #                     elif chord.quarter_duration == 4:
        #                         split = chord.split(1, 7)
        #                     elif chord.quarter_duration == 6:
        #                         split = chord.split(1, 11)
        #                     else:
        #                         pass
        #                     try:
        #                         split[1].to_rest()
        #                     except IndexError:
        #                         pass
        #                     new_chords.extend(split)
        #                 else:
        #                     new_chords.append(chord)
        #             beat._chords = new_chords

        self.score.write(path=result_path)
        # TestScore().assert_template(result_path=result_path)



from musicscore.dtd.dtd import ChildOccurrenceDTDConflict
from musicscore.musicxml.elements.xml_example import XMLExample, XMLExampleChild1, XMLExampleChild

from unittest import TestCase


class TestExample(TestCase):
    def setUp(self):
        self.example = XMLExample()

    def test_example(self):
        with self.assertRaises(ValueError):
            self.example.attribute_example = 2
        self.example.attribute_example = 'one'
        self.example.value = 2
        result = '''<example attribute-example="one">2</example>
'''
        self.assertEqual(self.example.to_string(), result)
        self.example.value = None
        self.example.add_child(XMLExampleChild())
        self.example.add_child(XMLExampleChild1())
        self.example.add_child(XMLExampleChild())
        result = '''<example attribute-example="one">
  <example-child-1/>
  <example-child-2/>
  <example-child-2/>
</example>
'''
        self.assertEqual(self.example.to_string(), result)
        with self.assertRaises(ChildOccurrenceDTDConflict):
            self.example.add_child(XMLExampleChild1())



from musicscore.dtd.dtd import Sequence, Choice, GroupReference, Element
from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.printobject import PrintObject
from musicscore.musicxml.attributes.printstyle import PrintStyle
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeFifths, TypeMode, TypeStep, TypeSemitones, TypeAccidentalValue, \
    TypeOctave


class KeyNumberAttribute(AttributeAbstract):
    def __init__(self, cancel=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('number', cancel, 'StaffNumber')


class Location(AttributeAbstract):
    """"""

    def __init__(self, location=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('location', location, 'TypeCancelLocation')


class ComplexTypeCancel(ComplexType, TypeFifths, Location):
    """
    A cancel element indicates that the old key signature should be cancelled before the new one appears. This will
    always happen when changing to C major or A minor and need not be specified then. The cancel value matches the
    fifths value of the cancelled key signature (e.g., a cancel of -2 will provide an explicit cancellation for changing
    from B flat major to F major). The optional location attribute indicates where the cancellation appears relative to
    the new key signature.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)


class Cancel(ComplexTypeCancel):
    """"""

    def __init__(self, value, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)


class Fifths(XMLElement, TypeFifths):
    """"""

    def __init__(self, value, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)


class Mode(XMLElement, TypeMode):
    """"""

    def __init__(self, value, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)


class KeyStep(XMLElement, TypeStep):
    """
    Non-traditional key signatures can be represented using the Humdrum/Scot concept of a list of altered tones.
    The key-step element indicates the pitch step to be altered, represented using the same names as in the step
    element.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)


class KeyAlter(XMLElement, TypeSemitones):
    """
    Non-traditional key signatures can be represented using the Humdrum/Scot concept of a list of altered tones.
    The key-step element indicates the pitch step to be altered, represented using the same names as in the step
    element.
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)


class ComplexKeyAccidental(ComplexType, TypeAccidentalValue):
    """
    Non-traditional key signatures can be represented using the Humdrum/Scot concept of a list of altered tones.
    The key-accidental element indicates the accidental to be displayed in the key signature, represented in the same
    manner as the accidental element. It is used for disambiguating microtonal accidentals.
    <xs:simpleContent>
        <xs:extension base="accidental-value">
            <xs:attribute name="smufl" type="smufl-accidental-glyph-name"/>
        </xs:extension>
    </xs:simpleContent>
    """

    def __init__(self, value, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)
        raise NotImplementedError()


class Number(AttributeAbstract):
    def __init__(self, number, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('number', number, 'PositiveInteger')


class CancelAttribute(AttributeAbstract):
    def __init__(self, cancel=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('cancel', cancel, 'TypeYesNo')


class ComplexTypeKeyOctave(ComplexType, TypeOctave, Number, CancelAttribute):
    """
    The key-octave element specifies in which octave an element of a key signature appears. The content specifies the
    octave value using the same values as the display-octave element. The number attribute is a positive integer that
    refers to the key signature element in left-to-right order. If the cancel attribute is set to yes, then this number
    refers to the canceling key signature specified by the cancel element in the parent key element. The cancel
    attribute cannot be set to yes if there is no corresponding cancel element within the parent key element. It is no
    by default
    """


class KeyOctave(ComplexTypeKeyOctave):

    def __init__(self, value, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)


"""
The traditional-key group represents a traditional key signature using the cycle of fifths.
"""
TraditionalKey = Sequence(
    Element(Cancel, min_occurrence=0),
    Element(Fifths),
    Element(Mode, min_occurrence=0)

)
"""
The non-traditional-key group represents a single alteration within a non-traditional key signature. A sequence of 
these groups makes up a non-traditional key signature
"""
NonTraditionalKey = Sequence(
    Element(KeyStep),
    Element(KeyAlter),
    Element(ComplexKeyAccidental, min_occurrence=0)
)


class TypeKey(ComplexType, KeyNumberAttribute, PrintStyle, PrintObject,
              OptionalUniqueId):
    """
    The key type represents a key signature. Both traditional and non-traditional key signatures are supported.
    The optional number attribute refers to staff numbers. If absent, the key signature applies to all staves in the
    part. Key signatures appear at the start of each system unless the print-object attribute has been set to "no".
    """
    _DTD = (
        Sequence(
            Choice(
                GroupReference(TraditionalKey),
                GroupReference(NonTraditionalKey, min_occurrence=0, max_occurrence=None)
            ),
            Element(KeyOctave, min_occurrence=0, max_occurrence=None)

        )
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from unittest import TestCase
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
import os

from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class TestAddMeasure(TestCase):
    def setUp(self):
        self.score = TreeScoreTimewise()
        self.score.add_part()

    def test_1(self):
        self.score.add_measure().time = (3, 4)
        self.score.add_measure().time = (2, 4)
        self.score.add_measure().time = (2, 4)
        self.score.add_measure().time = (2, 4)
        self.score.get_measure(4).time.force_show = True
        self.score.add_measure().time = (3, 8)

        self.score.finish()
        result_path = path + '_test_1'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        xml_path = path + '_test_1.xml'

        sf = SimpleFormat(quarter_durations=5 * [2.25])
        sf.chords[0].add_bracket(type='start', line_end='none', relative_x=50)
        sf.chords[1].add_bracket(type='continue', line_end='none')
        sf.chords[2].add_bracket(type='stop', line_end='arrow', relative_x=-30, relative_y=20)

        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        TestScore().assert_template(xml_path)



from unittest import TestCase

from musicscore.dtd.dtd import ChildIsNotOptional
from musicscore.musicxml.groups.layout import PageLayout
from musicscore.musicxml.groups.margins import LeftMargin, RightMargin
from musicscore.musicxml.types.complextypes.pagelayout import PageMargins


class Test(TestCase):
    def test_1(self):
        page_margins = PageMargins()
        page_margins.add_child(LeftMargin(70))
        page_margins.add_child(RightMargin(70))
        with self.assertRaises(ChildIsNotOptional):
            page_margins.to_string()

    def test_2(self):
        page_layout = PageLayout()
        page_margins = page_layout.add_child(PageMargins())
        page_margins.add_child(LeftMargin(70))
        page_margins.add_child(RightMargin(70))
        with self.assertRaises(ChildIsNotOptional):
            page_layout.to_string()



from musicscore.dtd.dtd import Sequence, GroupReference, Element
from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.attributes.barline import BarlineAttributes
from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.groups.common import Editorial
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.complextypes.segno import ComplexTypeSegno
from musicscore.musicxml.types.complextypes.wavyline import ComplexTypeWavyLine
from musicscore.musicxml.types.simple_type import TypeBarStyle, TypeBackwardForward
from musicscore.musicxml.elements.xml_element import XMLElement


class WavyLine(ComplexTypeWavyLine):
    _TAG = 'wavy-line'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class Segno(ComplexTypeSegno):
    def __init__(self, *args, **kwargs):
        super().__init__(tag='segno', *args, **kwargs)


class Coda(XMLElement):
    def __init__(self, *args, **kwargs):
        super().__init__(tag='coda', *args, **kwargs)
        raise NotImplementedError()


class Fermata(XMLElement):
    def __init__(self, *args, **kwargs):
        super().__init__(tag='fermata', *args, **kwargs)
        raise NotImplementedError('Fermata')


class Ending(XMLElement):
    def __init__(self, *args, **kwargs):
        super().__init__(tag='ending', *args, **kwargs)
        raise NotImplementedError('Ending')


class Times(AttributeAbstract):

    def __init__(self, times=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('times', times, 'NonNegativeInteger')


class Winged(AttributeAbstract):

    def __init__(self, winged=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('winged', winged, 'TypeWinged')


class ComplexTypeRepeat(ComplexType, Times, Winged):
    """The repeat type represents repeat marks. The start of the repeat has a forward direction while the end of the
    repeat has a backward direction. Backward repeats that are not part of an ending can use the times attribute to
    indicate the number of times the repeated section is played.
    """

    def __init__(self, tag, direction, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.direction = direction

    @property
    def direction(self):
        return self.get_attribute('direction')

    @direction.setter
    def direction(self, value):
        if value is None:
            self.remove_attribute('direction')
        else:
            TypeBackwardForward(value)
            self._ATTRIBUTES.insert(0, 'direction')
            self.set_attribute('direction', value)


class Repeat(ComplexTypeRepeat):
    def __init__(self, *args, **kwargs):
        super().__init__(tag='repeat', *args, **kwargs)


'''
	<xs:complexType name="barline">
		<xs:sequence>
			<xs:element name="bar-style" type="bar-style-color" minOccurs="0"/>
			<xs:group ref="editorial"/>
			<xs:element name="wavy-line" type="wavy-line" minOccurs="0"/>
			<xs:element name="segno" type="segno" minOccurs="0"/>
			<xs:element name="coda" type="coda" minOccurs="0"/>
			<xs:element name="fermata" type="fermata" minOccurs="0" maxOccurs="2"/>
			<xs:element name="ending" type="ending" minOccurs="0"/>
			<xs:element name="repeat" type="repeat" minOccurs="0"/>
		</xs:sequence>
		<xs:attribute name="location" type="right-left-middle" default="right"/>
		<xs:attribute name="segno" type="xs:token"/>
		<xs:attribute name="coda" type="xs:token"/>
		<xs:attribute name="divisions" type="divisions"/>
		<xs:attributeGroup ref="optional-unique-id"/>
	</xs:complexType>
'''


class ComplexTypeBarStyleColor(ComplexType, TypeBarStyle, Color):
    """
    The bar-style-color type contains barline style and color information.
    """

    def __init__(self, tag, value, *args, **kwargs):
        super().__init__(tag=tag, value=value, *args, **kwargs)


class BarStyle(ComplexTypeBarStyleColor):
    """"""

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag='bar-style', value=value, *args, **kwargs)


class Barline(ComplexType, BarlineAttributes):
    """
    If a barline is other than a normal single barline, it should be represented by a barline type that describes it.
    This includes information about repeats and multiple endings, as well as line style. Barline data is on the same
    level as the other musical data in a score - a child of a measure in a partwise score, or a part in a timewise
    score. This allows for barlines within measures, as in dotted barlines that subdivide measures in complex meters.
    The two fermata elements allow for fermatas on both sides of the barline (the lower one inverted).

    Barlines have a location attribute to make it easier to process barlines independently of the other musical data in
    a score. It is often easier to set up measures separately from entering notes. The location attribute must match
    where the barline element occurs within the rest of the musical data in the score. If location is left, it should be
    the first element in the measure, aside from the print, bookmark, and link elements. If location is right, it should
    be the last element, again with the possible exception of the print, bookmark, and link elements. If no location is
    specified, the right barline is the default. The segno, coda, and divisions attributes work the same way as in the
    sound element. They are used for playback when barline elements contain segno or coda child elements.
    """
    _DTD = Sequence(
        Element(BarStyle, min_occurrence=0),
        GroupReference(Editorial),
        Element(WavyLine, min_occurrence=0),
        Element(Segno, min_occurrence=0),
        Element(Coda, min_occurrence=0),
        Element(Fermata, min_occurrence=0, max_occurrence=2),
        Element(Ending, min_occurrence=0),
        Element(Repeat, min_occurrence=0)
    )

    def __init__(self, *args, **kwargs):
        super().__init__(tag='barline', *args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract, TypeLeftCenterRight


class Justify(AttributeAbstract):
    """
    The justify attribute is used to indicate left, center, or right justification. The default value varies for
    different elements. For elements where the justify attribute is present but the halign attribute is not, the justify
    attribute indicates horizontal alignment as well as justification.
    """

    def __init__(self, justify=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('justify', justify, "TypeLeftCenterRight")
        TypeLeftCenterRight
